import re
from re import Pattern
from typing import Any, List, Optional

from unidecode import unidecode

from tempren.path_generator import File
from tempren.template.tree_elements import Tag


class UnidecodeTag(Tag):
    """Replace special unicode characters in context with ASCII equivalents"""

    require_context = True

    def process(self, file: File, context: Optional[str]) -> str:
        assert context is not None
        return unidecode(context)


class RemoveTag(Tag):
    """Remove parts of the context based on RegEx patterns"""

    require_context = True
    patterns: List[Pattern]

    def configure(self, *patterns: str, ignore_case: bool = False):  # type: ignore
        flags = 0
        if ignore_case:
            flags |= re.IGNORECASE  # TODO: make this configurable?
        self.patterns = list(map(lambda p: re.compile(p, flags), patterns))

    def process(self, file: File, context: Optional[str]) -> str:
        assert context is not None
        result = context
        for pattern in self.patterns:
            result = pattern.sub("", result)
        return result


class ReplaceTag(Tag):
    """Replaces regex pattern with specified replacement"""

    require_context = True
    pattern: Pattern
    replacement: str

    def configure(self, pattern: str, replacement: str):  # type: ignore
        # TODO: add ignore-case option?
        self.pattern = re.compile(pattern)
        self.replacement = replacement

    def process(self, file: File, context: Optional[str]) -> str:
        assert context
        return self.pattern.sub(self.replacement, context)


class CollapseTag(Tag):
    """Collapse specified repeating characters in context"""

    require_context = True
    pattern: Pattern

    def configure(self, characters: str = " "):  # type: ignore
        # TODO: check characters for empty string?
        self.pattern = re.compile(f"(?<=[{characters}])[{characters}]+")

    def process(self, file: File, context: Optional[str]) -> str:
        assert context is not None
        return self.pattern.sub("", context)


class UpperTag(Tag):
    """Makes context uppercase"""

    require_context = True

    def process(self, file: File, context: Optional[str]) -> str:
        assert context is not None
        return context.upper()


class LowerTag(Tag):
    """Makes context lowercase"""

    require_context = True

    def process(self, file: File, context: Optional[str]) -> str:
        assert context is not None
        return context.lower()


class StripTag(Tag):
    """Removes dangling characters on the ends of provided context"""

    require_context = True
    strip_characters: str = " "
    left: bool = False
    right: bool = False

    def configure(self, strip_characters: str = strip_characters, left: bool = False, right: bool = False):  # type: ignore
        self.strip_characters = strip_characters
        self.left = left
        self.right = right

    def process(self, file: File, context: Optional[str]) -> str:
        assert context is not None
        if self.left and not self.right:
            return context.lstrip(self.strip_characters)
        if self.right and not self.left:
            return context.rstrip(self.strip_characters)
        return context.strip(self.strip_characters)


class TrimTag(Tag):
    """Trims context to a specified width by cropping left/right side off

    If width is less than 0, trims abs(width) number of characters from the side context instead.
    """

    require_context = True
    width: int
    left: bool = False
    right: bool = False

    def configure(self, width: int, left: bool = False, right: bool = False):  # type: ignore
        assert (
            width != 0
        ), "specify positive width or negative number of characters to trim"
        self.width = width
        assert (
            not left or not right
        ), "left and right cannot be specified at the same time"
        assert left or right, "no trim direction specified - use left or right"
        self.left = left
        self.right = right

    def process(self, file: File, context: Optional[str]) -> str:
        assert context is not None
        if self.left:
            return context[-self.width :]
        return context[: self.width]


class CapitalizeTag(Tag):
    """Capitalizes first letter of the context"""

    require_context = True

    def process(self, file: File, context: Optional[str]) -> str:
        assert context is not None
        return context.capitalize()


class TitleTag(Tag):
    """Capitalizes first letter of every word in the context"""

    require_context = True

    def process(self, file: File, context: Optional[str]) -> str:
        assert context is not None
        return context.title()


class PadTag(Tag):
    """Adds padding to the context to match specified width"""

    require_context = True
    width: int
    character: str = " "
    left: bool = False
    right: bool = False
    center: bool = False

    def configure(self, width: int, character: str = character, left: bool = False, right: bool = False):  # type: ignore
        assert width > 0, "width have to be positive integer"
        self.width = width
        assert len(character) == 1, "single character have to be provided for padding"
        self.character = character
        assert left or right, "'left' or 'right' direction have to be provided"
        self.left = left
        self.right = right

    def process(self, file: File, context: Optional[str]) -> str:
        assert context is not None
        if len(context) >= self.width:
            return context

        if self.left and self.right:
            return context.center(self.width, self.character)
        if self.left:
            return context.rjust(self.width, self.character)
        else:
            return context.ljust(self.width, self.character)


class SplitCaseTag(Tag):
    """Split text into words on the case boundary"""

    require_context = True
    separator: str
    _pattern = re.compile(r"([a-z])([A-Z])")

    def configure(self, separator: str = " "):  # type: ignore
        """
        :param separator: character/text used to join resulting words
        """
        assert separator
        self.separator = separator

    def process(self, file: File, context: Optional[str]) -> Any:
        assert context is not None
        return self._pattern.sub("".join((r"\1", self.separator, r"\2")), context)



import importlib
import inspect
import logging
import pkgutil
from dataclasses import dataclass
from functools import reduce
from logging import Logger
from types import ModuleType
from typing import Dict, List, Mapping, Optional, Tuple, Type, Union

from antlr4 import CommonTokenStream, InputStream  # type: ignore
from antlr4.error.ErrorListener import ErrorListener  # type: ignore

from .grammar.TagTemplateLexer import TagTemplateLexer
from .grammar.TagTemplateParser import TagTemplateParser
from .grammar.TagTemplateParserVisitor import TagTemplateParserVisitor
from .tree_elements import (
    Location,
    Pattern,
    PatternElement,
    RawText,
    Tag,
    TagFactory,
    TagFactoryFromClass,
    TagInstance,
    TagName,
    TagPlaceholder,
)

ArgValue = Union[str, int, bool]


escaped_characters = ("'", "\\", "{", "}", "|")
replacements = list(("\\" + ec, ec) for ec in escaped_characters)


def unescape(text: str) -> str:
    return reduce(
        lambda acc, replacement: acc.replace(*replacement), replacements, text
    )


IGNORED_TERMINAL = object()
ARGUMENT_VALUE_TYPES = (
    TagTemplateLexer.NUMERIC_VALUE,
    TagTemplateLexer.BOOLEAN_VALUE,
    TagTemplateLexer.STRING_VALUE,
)


def location_from_symbol(symbol) -> Location:
    return Location(
        line=symbol.line, column=symbol.start, length=symbol.stop - symbol.start + 1
    )


def merge_locations(first: Optional[Location], second: Location) -> Location:
    if not first:
        return second
    assert first.line == second.line
    return Location(
        first.line, first.column, second.column + second.length - first.column
    )


class _TreeVisitor(TagTemplateParserVisitor):
    def defaultResult(self) -> List[PatternElement]:
        return list()

    def visitTerminal(self, node):
        if node.getSymbol().type not in ARGUMENT_VALUE_TYPES:
            return IGNORED_TERMINAL
        raise NotImplementedError()

    def aggregateResult(
        self, pattern: List[PatternElement], element: PatternElement
    ) -> List[PatternElement]:
        if element is IGNORED_TERMINAL:
            return pattern
        return pattern + [element]

    def visitRootPattern(self, ctx: TagTemplateParser.RootPatternContext) -> Pattern:
        return self.visitPattern(ctx.pattern())

    def visitPipeList(
        self, ctx: TagTemplateParser.PipeListContext
    ) -> List[TagPlaceholder]:
        if ctx.errorNonTagInPipeList:
            non_tag_symbol = ctx.errorNonTagInPipeList.children[0].symbol
            raise TemplateSyntaxError(
                message=f"non-tag in the pipe list"
            ).with_location(location_from_symbol(non_tag_symbol))
        tag_list = self.visitChildren(ctx)
        return list(filter(bool, tag_list))

    def visitPattern(self, ctx: TagTemplateParser.PatternContext) -> Pattern:
        pattern_elements = self.visitChildren(ctx)
        pipe_list = ctx.pipeList()
        if pipe_list:
            pipe_tags = self.visitPipeList(pipe_list)
            context = Pattern(pattern_elements[:-1])
            for tag_placeholder in pipe_tags:
                tag_placeholder.context = context
                context = Pattern([tag_placeholder])
            return context
        return Pattern(pattern_elements)

    def visitTag(self, ctx: TagTemplateParser.TagContext) -> TagPlaceholder:
        category_name = None
        if ctx.categoryId:
            category_name = ctx.categoryId.text
        if ctx.errorMissingTagId:
            raise TemplateSyntaxError(message=f"missing tag name").with_location(
                location_from_symbol(ctx.errorMissingTagId)
            )
        if ctx.errorMissingCategoryId:
            raise TemplateSyntaxError(message=f"missing category name").with_location(
                location_from_symbol(ctx.errorMissingCategoryId)
            )
        if ctx.errorNoArgumentList:
            raise TemplateSyntaxError(
                message=f"missing argument list for tag '{ctx.errorNoArgumentList.text}'"
            ).with_location(location_from_symbol(ctx.errorNoArgumentList))
        tag_name: str = ctx.tagId.text  # type: ignore
        if ctx.errorUnclosedContext:
            raise TemplateSyntaxError(
                message=f"missing closing context bracket for tag '{tag_name}'"
            ).with_location(location_from_symbol(ctx.errorUnclosedContext))
        args, kwargs = self.visitArgumentList(ctx.argumentList())
        context = None
        context_pattern = ctx.pattern()
        if context_pattern is not None:
            context = self.visitPattern(ctx.pattern())

        tag = TagPlaceholder(
            tag_name=TagName(tag_name, category_name),
            args=args,
            kwargs=kwargs,
            context=context,
        )

        tag_name_location = location_from_symbol(ctx.tagId)
        if ctx.categoryId:
            tag_category_location = location_from_symbol(ctx.categoryId)
            tag.location = merge_locations(tag_category_location, tag_name_location)
        else:
            tag.location = tag_name_location
        return tag

    def visitArgumentList(
        self, ctx: TagTemplateParser.ArgumentListContext
    ) -> Tuple[List[ArgValue], Mapping[str, ArgValue]]:
        if ctx.errorUnclosedArgumentList:
            raise TemplateSyntaxError(
                message="missing closing argument list bracket"
            ).with_location(location_from_symbol(ctx.errorUnclosedArgumentList))
        collected_arguments = super().visitArgumentList(ctx)
        args = [
            arg_val for arg_name, arg_val in collected_arguments if arg_name is None
        ]
        kwargs = {
            arg_name: arg_val
            for arg_name, arg_val in collected_arguments
            if arg_name is not None
        }
        return args, kwargs

    def visitArgumentValue(
        self, ctx: TagTemplateParser.ArgumentValueContext
    ) -> ArgValue:
        if ctx.BOOLEAN_VALUE():
            return ctx.BOOLEAN_VALUE().getText().lower() == "true"
        elif ctx.NUMERIC_VALUE():
            return int(ctx.NUMERIC_VALUE().getText())
        elif ctx.STRING_VALUE():
            str_val = ctx.STRING_VALUE().getText()
            assert len(str_val) >= 2
            str_val = str_val[1:-1]
            return unescape(str_val)
        raise NotImplementedError("Unknown argument value token: " + ctx.getText())

    def visitArgument(
        self, ctx: TagTemplateParser.ArgumentContext
    ) -> Tuple[Optional[str], ArgValue]:
        arg_name = ctx.ARG_NAME().getText() if ctx.ARG_NAME() else None
        if ctx.argumentValue():
            arg_value = self.visitArgumentValue(ctx.argumentValue())
        else:
            arg_value = True
        return arg_name, arg_value

    def visitRawText(self, ctx: TagTemplateParser.RawTextContext) -> RawText:
        symbol = ctx.TEXT().getSymbol()
        # CHECK: is location necessary for RawText?
        raw_text = RawText(text=unescape(ctx.TEXT().getText()))
        raw_text.location = location_from_symbol(symbol)
        return raw_text


class TagTreeBuilder:
    log: logging.Logger

    def __init__(self):
        self.log = logging.getLogger(self.__class__.__name__)

    def parse(self, text: str) -> Pattern:
        self.log.debug("Parsing '%s'", text)
        lexer = TagTemplateLexer(InputStream(text))
        token_stream = CommonTokenStream(lexer)
        token_stream.fill()
        parser = TagTemplateParser(token_stream)
        parser.addErrorListener(TagTemplateErrorListener())

        visitor = _TreeVisitor()
        root_pattern = visitor.visitRootPattern(parser.rootPattern())
        root_pattern.source_representation = text
        return root_pattern


class TagTemplateErrorListener(ErrorListener):
    def syntaxError(self, recognizer, offendingSymbol, line, column, msg, e):
        error_location = Location(line, column, len(offendingSymbol.text))
        if "extraneous input" in msg or "mismatched input" in msg:
            raise TemplateSyntaxError(
                f"unexpected symbol '{offendingSymbol.text}'"
            ).with_location(error_location)
        else:
            raise TemplateSyntaxError(msg).with_location(error_location)


class TemplateError(Exception):
    """Represents an error in the template itself"""

    message: str
    location: Location
    template: str

    def __init__(self, message: str):
        self.message = message
        self.location = Location(0, 0, 0)

    def with_location(self, location: Location) -> "TemplateError":
        self.location = location
        return self

    def __str__(self) -> str:
        if self.location:
            return f"{self.location}: {self.message}"
        else:
            return self.message


class TemplateSyntaxError(TemplateError):
    pass


class TemplateSemanticError(TemplateError):
    pass


class TagError(TemplateSemanticError):
    tag_name: TagName

    def __init__(self, tag_name: TagName, message: str):
        assert tag_name
        self.tag_name = tag_name
        super().__init__(f"Error in tag '{self.tag_name}': {message}")


class UnknownTagError(TagError):
    def __init__(self, tag_name: TagName):
        super().__init__(tag_name, f"Unknown tag name: {tag_name.name}")

    def with_location(self, whole_name_location: Location) -> "TemplateError":
        if not self.tag_name.category:
            return super().with_location(whole_name_location)
        name_location = Location(
            whole_name_location.line,
            whole_name_location.column + len(self.tag_name.category) + 1,
            len(self.tag_name.name),
        )
        return super().with_location(name_location)


class UnknownCategoryError(TagError):
    def __init__(self, tag_name: TagName):
        super().__init__(tag_name, f"Unknown category name: {tag_name.category}")

    def with_location(self, whole_name_location: Location) -> "TemplateError":
        assert self.tag_name.category
        category_location = Location(
            whole_name_location.line,
            whole_name_location.column,
            len(self.tag_name.category),
        )
        return super().with_location(category_location)


class AmbiguousTagError(TagError):
    category_names: List[str]

    def __init__(self, tag_name: TagName, category_names: List[str]):
        super().__init__(
            tag_name,
            f"This tag name is present in multiple categories: {', '.join(category_names)}",
        )
        self.category_names = category_names


class ContextMissingError(TagError):
    def __init__(self, tag_name: TagName):
        super().__init__(tag_name, f"Context is required for this tag")


class ContextForbiddenError(TagError):
    def __init__(self, tag_name: TagName):
        super().__init__(tag_name, f"This tag cannot be used with context")


class ConfigurationError(TagError):
    def __init__(self, tag_name: TagName, message: str):
        super().__init__(tag_name, f"Configuration not valid: {message}")


class TagCategory:
    log: Logger

    name: str
    description: Optional[str] = None
    tag_map: Dict[str, TagFactory]

    def __init__(self, name: str, description: Optional[str] = None):
        self.log = logging.getLogger(self.__class__.__name__)
        self.tag_map = {}
        self.name = name
        self.description = description

    def register_tag(self, tag_class: Type[Tag], tag_name: Optional[str] = None):
        _simple_tag_factory = TagFactoryFromClass(tag_class, tag_name)
        self.log.debug(
            f"Registering class {tag_class} as {_simple_tag_factory.tag_name} tag"
        )
        self.register_tag_factory(_simple_tag_factory, tag_name)

    def register_tag_factory(
        self, tag_factory: TagFactory, tag_name: Optional[str] = None
    ):
        if tag_name is None:
            tag_name = tag_factory.tag_name
        if not tag_name:
            raise ValueError(f"Invalid tag name '{repr(tag_name)}'")
        if tag_name in self.tag_map:
            raise ValueError(f"Factory for tag '{tag_name}' already registered")
        self.tag_map[tag_name] = tag_factory

    def find_tag_factory(self, tag_name: str) -> Optional[TagFactory]:
        return self.tag_map.get(tag_name, None)


class TagRegistry:
    log: Logger
    _tag_class_suffix = "Tag"
    category_map: Dict[str, TagCategory]

    def __init__(self):
        self.log = logging.getLogger(self.__class__.__name__)
        self.category_map = {}

    @property
    def categories(self) -> List[str]:
        return sorted(self.category_map.keys())

    def find_category(self, category_name: str) -> Optional[TagCategory]:
        category = self.category_map.get(category_name, None)
        if category:
            return category
        category_name = category_name.lower()
        return self.category_map.get(category_name, None)

    def get_tag_factory(self, tag_name: TagName) -> TagFactory:
        if tag_name.category is None:
            return self._get_tag_factory_by_unique_name(tag_name)
        else:
            return self._get_tag_factory_by_name_and_category(tag_name)

    def _get_tag_factory_by_unique_name(self, tag_name: TagName) -> TagFactory:
        # In case there are tags with the same name in multiple categories,
        # category name have to be specified explicitly
        found_tag_factories: Dict[str, TagFactory] = {}
        for category in self.category_map.values():
            tag_factory = category.find_tag_factory(tag_name.name)
            if tag_factory:
                found_tag_factories[category.name] = tag_factory

        if not found_tag_factories:
            raise UnknownTagError(tag_name)
        elif len(found_tag_factories) > 1:
            category_names = sorted(list(found_tag_factories.keys()))
            raise AmbiguousTagError(tag_name, category_names)

        return next(iter(found_tag_factories.values()))

    def _get_tag_factory_by_name_and_category(self, tag_name: TagName) -> TagFactory:
        assert tag_name.category
        tag_category = self.find_category(tag_name.category)
        if tag_category is None:
            raise UnknownCategoryError(tag_name)

        tag_factory = tag_category.find_tag_factory(tag_name.name)
        if tag_factory is None:
            raise UnknownTagError(tag_name)

        return tag_factory

    def bind(self, pattern: Pattern) -> Pattern:
        return self._rewrite_pattern(pattern)

    def _rewrite_pattern(self, pattern: Pattern) -> Pattern:
        new_elements: List[PatternElement] = []
        for element in pattern.sub_elements:
            if isinstance(element, TagPlaceholder):
                new_elements.append(self._rewrite_tag_placeholder(element))
            else:
                new_elements.append(element)
        bound_pattern = Pattern(new_elements)
        bound_pattern.source_representation = pattern.source_representation
        return bound_pattern

    def _rewrite_tag_placeholder(self, tag_placeholder: TagPlaceholder) -> TagInstance:
        try:
            tag_factory = self.get_tag_factory(tag_placeholder.tag_name)

            self.log.debug(
                "Creating tag '%s' with arguments: %s %s",
                tag_placeholder.tag_name,
                tag_placeholder.args,
                tag_placeholder.kwargs,
            )
            tag = tag_factory(*tag_placeholder.args, **tag_placeholder.kwargs)
        except TemplateError as template_error:
            raise template_error.with_location(tag_placeholder.location)
        except Exception as exc:
            raise ConfigurationError(tag_placeholder.tag_name, str(exc)).with_location(
                tag_placeholder.location
            ) from exc

        if tag.require_context is not None:
            if tag_placeholder.context and not tag.require_context:
                raise ContextForbiddenError(tag_placeholder.tag_name).with_location(
                    tag_placeholder.location
                )
            elif tag_placeholder.context is None and tag.require_context:
                raise ContextMissingError(tag_placeholder.tag_name).with_location(
                    tag_placeholder.location
                )

        context_pattern: Optional[Pattern] = None
        if tag_placeholder.context:
            context_pattern = self._rewrite_pattern(tag_placeholder.context)

        return TagInstance(tag, context=context_pattern)

    def register_category(
        self, category_name: str, description: Optional[str] = None
    ) -> TagCategory:
        # TODO: This method should receive already build (non-empty) TagCategory
        if self.find_category(category_name) is not None:
            raise ValueError(f"Category '{category_name}' already registered")
        new_category = TagCategory(category_name)
        self.category_map[category_name] = new_category
        return new_category

    def register_tags_in_module(self, module: ModuleType):
        self.log.debug(f"Discovering tags in module '{module}'")

        if module.__package__:
            category_name = module.__name__[len(module.__package__) + 1 :]
        else:
            category_name = module.__name__

        def is_tag_class(klass: type):
            if (
                not inspect.isclass(klass)
                or not issubclass(klass, Tag)
                or inspect.isabstract(klass)
                or klass == Tag
            ):
                return False
            return klass.__name__.endswith(self._tag_class_suffix)

        # TODO: do not register empty modules
        module_category = self.register_category(category_name)
        for _, tag_class in inspect.getmembers(module, is_tag_class):
            module_category.register_tag(tag_class)

    def register_tags_in_package(self, package):
        self.log.debug(f"Discovering tags in package '{package}'")

        for _, name, is_pkg in pkgutil.walk_packages(
            package.__path__, package.__name__ + "."
        ):
            if is_pkg:
                continue
            try:
                self.log.debug(f"Trying to load {name} module")
                module = importlib.import_module(name)
                self.register_tags_in_module(module)
            except NotImplementedError as exc:
                self.log.warning(f"Module {name} is currently unsupported: {exc}")
            except Exception as exc:
                self.log.error(exc, f"Could not load module {name}")



from abc import ABC, abstractmethod
from datetime import timedelta
from fractions import Fraction
from typing import Any, Optional

from pymediainfo import MediaInfo

from tempren.path_generator import File
from tempren.template.tree_elements import (
    FileNotSupportedError,
    MissingMetadataError,
    Tag,
)

if not MediaInfo.can_parse():
    raise NotImplementedError("MediaInfo library not found")


class MediaInfoTagBase(Tag, ABC):
    require_context = False

    def process(self, file: File, context: Optional[str]) -> Any:
        media_info = MediaInfo.parse(file.absolute_path)
        if len(media_info.tracks) < 2:  # General track seems always present
            raise FileNotSupportedError()
        return self.extract_metadata(media_info)

    @abstractmethod
    def extract_metadata(self, media_info: MediaInfo) -> Any:
        raise NotImplementedError()


class VideoInfoTagBase(MediaInfoTagBase, ABC):
    def extract_metadata(self, media_info: MediaInfo) -> Any:
        if not media_info.video_tracks:
            raise MissingMetadataError()
        return self.extract_video_metadata(media_info.video_tracks[0])

    @abstractmethod
    def extract_video_metadata(self, video_track) -> Any:
        raise NotImplementedError()


class WidthTag(VideoInfoTagBase):
    """Video width in pixels"""

    def extract_video_metadata(self, video_track) -> Any:
        return video_track.width


class HeightTag(VideoInfoTagBase):
    """Video height in pixels"""

    def extract_video_metadata(self, video_track) -> Any:
        return video_track.height


class AspectRatioTag(VideoInfoTagBase):
    """Video aspect ratio (in fractional W:H or decimal format)"""

    use_decimal: bool

    def configure(self, decimal: bool = False):
        """
        :param decimal: use decimal notation
        """
        self.use_decimal = decimal

    def extract_video_metadata(self, video_track) -> Any:
        if self.use_decimal:
            return video_track.width / video_track.height
        else:
            aspect_ratio = Fraction(video_track.width, video_track.height)
            return f"{aspect_ratio.numerator}:{aspect_ratio.denominator}"


class FrameRateTag(VideoInfoTagBase):
    """Decimal frame rate per second"""

    def extract_video_metadata(self, video_track) -> Any:
        return float(video_track.frame_rate)


class VideoCodecTag(VideoInfoTagBase):
    """Name of video encoding codec"""

    def extract_video_metadata(self, video_track) -> Any:
        return video_track.format


class FrameCountTag(VideoInfoTagBase):
    """Number of frames in the file"""

    def extract_video_metadata(self, video_track) -> Any:
        return int(video_track.frame_count)


class DurationTag(VideoInfoTagBase):
    """Video duration in seconds"""

    def extract_video_metadata(self, video_track) -> timedelta:
        return timedelta(milliseconds=video_track.duration)


class BitRateTag(VideoInfoTagBase):
    """Video stream bit rate"""

    def extract_video_metadata(self, video_track) -> int:
        return int(video_track.bit_rate)



# Generated from TagTemplateParser.g4 by ANTLR 4.10.1
from antlr4 import *

if __name__ is not None and "." in __name__:
    from .TagTemplateParser import TagTemplateParser
else:
    from TagTemplateParser import TagTemplateParser

# This class defines a complete generic visitor for a parse tree produced by TagTemplateParser.


class TagTemplateParserVisitor(ParseTreeVisitor):

    # Visit a parse tree produced by TagTemplateParser#rootPattern.
    def visitRootPattern(self, ctx: TagTemplateParser.RootPatternContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by TagTemplateParser#tag.
    def visitTag(self, ctx: TagTemplateParser.TagContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by TagTemplateParser#pipeList.
    def visitPipeList(self, ctx: TagTemplateParser.PipeListContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by TagTemplateParser#pattern.
    def visitPattern(self, ctx: TagTemplateParser.PatternContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by TagTemplateParser#argumentList.
    def visitArgumentList(self, ctx: TagTemplateParser.ArgumentListContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by TagTemplateParser#argument.
    def visitArgument(self, ctx: TagTemplateParser.ArgumentContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by TagTemplateParser#argumentValue.
    def visitArgumentValue(self, ctx: TagTemplateParser.ArgumentValueContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by TagTemplateParser#rawText.
    def visitRawText(self, ctx: TagTemplateParser.RawTextContext):
        return self.visitChildren(ctx)


del TagTemplateParser



import logging
from abc import ABC
from pathlib import Path

from tempren.path_generator import File, InvalidFilenameError, PathGenerator
from tempren.template.tree_elements import Pattern


class TemplateGenerator(PathGenerator, ABC):
    log: logging.Logger
    pattern: Pattern

    def __init__(self, pattern: Pattern):
        self.log = logging.getLogger(__name__)
        self.log.debug("Creating template generator with template: %s", pattern)
        self.pattern = pattern

    def generate_replacement(self, file: File) -> str:
        self.log.debug("Rendering template for '%s'", file.relative_path)
        rendered_template = self.pattern.process(file)
        self.log.debug("Rendered template: '%s'", rendered_template)
        return rendered_template


class TemplateNameGenerator(TemplateGenerator):
    def generate(self, file: File) -> Path:
        new_name = self.generate_replacement(file)
        try:
            return file.relative_path.with_name(new_name)
        except ValueError:
            raise InvalidFilenameError(new_name)


class TemplatePathGenerator(TemplateGenerator):
    def generate(self, file: File) -> Path:
        return Path(self.generate_replacement(file))



import datetime
import os.path
from typing import Any, Optional

from tempren.path_generator import File
from tempren.template.tree_elements import Tag


class SizeTag(Tag):
    """File size in bytes"""

    require_context = False

    def process(self, file: File, context: Optional[str]) -> int:
        assert context is None
        return os.path.getsize(file.absolute_path)


class MTimeTag(Tag):
    """File modification time (in ISO 8601 format)"""

    require_context = False

    def process(self, file: File, context: Optional[str]) -> Any:
        assert context is None
        mtime_seconds = os.path.getmtime(file.absolute_path)
        mtime = datetime.datetime.fromtimestamp(mtime_seconds)
        return mtime.isoformat()


class OwnerTag(Tag):
    """Name of the user owning processed file"""

    require_context = False

    def process(self, file: File, context: Optional[str]) -> str:
        # TODO: Maybe if context is present parse it as a path?
        return file.absolute_path.owner()


class GroupTag(Tag):
    """Name of the group owning processed file"""

    require_context = False

    def process(self, file: File, context: Optional[str]) -> str:
        # TODO: Maybe if context is present parse it as a path?
        return file.absolute_path.group()



from abc import ABC, abstractmethod
from dataclasses import dataclass
from pathlib import Path, PosixPath
from typing import Any


@dataclass
class File:
    input_directory: Path
    relative_path: Path

    @classmethod
    def from_path(cls, path_representation: str) -> "File":
        path = Path(path_representation)
        return File(path.parent, Path(path.name))

    @property
    def absolute_path(self) -> Path:
        return self.input_directory / self.relative_path

    def __init__(self, input_directory: Path, relative_path: Path):
        assert input_directory.is_absolute()
        assert not relative_path.is_absolute()
        self.input_directory = input_directory
        self.relative_path = relative_path

    def __str__(self):
        return str(self.relative_path)

    def __repr__(self):
        return repr(str(self))


class InvalidFilenameError(Exception):
    generated_name: str

    def __init__(self, invalid_name: str):
        super().__init__(f"Invalid name: '{invalid_name}'")
        self.generated_name = invalid_name


class PathGenerator(ABC):
    @abstractmethod
    def generate(self, file: File) -> Path:
        raise NotImplementedError()


class ExpressionEvaluationError(Exception):
    expression: str

    @property
    def message(self) -> str:
        return str(self.__cause__)

    def __init__(self, expression: str):
        self.expression = expression


class TemplateEvaluationError(Exception):
    file: File
    template: str
    expression: str
    message: str

    def __init__(self, file: File, template: str, expression: str, message: str):
        self.file = file
        self.template = template
        self.expression = expression
        self.message = message


_evaluation_locals = {"PosixPath": PosixPath}


def evaluate_expression(python_expression: str) -> Any:
    try:
        evaluation_result = eval(python_expression, {}, _evaluation_locals)
        return evaluation_result
    except Exception as exc:
        raise ExpressionEvaluationError(python_expression) from exc



# Generated from TagTemplateParser.g4 by ANTLR 4.10.1
# encoding: utf-8
import sys
from io import StringIO

from antlr4 import *

if sys.version_info[1] > 5:
    from typing import TextIO
else:
    from typing.io import TextIO


def serializedATN():
    return [
        4,
        1,
        19,
        109,
        2,
        0,
        7,
        0,
        2,
        1,
        7,
        1,
        2,
        2,
        7,
        2,
        2,
        3,
        7,
        3,
        2,
        4,
        7,
        4,
        2,
        5,
        7,
        5,
        2,
        6,
        7,
        6,
        2,
        7,
        7,
        7,
        1,
        0,
        1,
        0,
        1,
        0,
        1,
        1,
        1,
        1,
        1,
        1,
        3,
        1,
        23,
        8,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        3,
        1,
        31,
        8,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        3,
        1,
        47,
        8,
        1,
        1,
        2,
        1,
        2,
        4,
        2,
        51,
        8,
        2,
        11,
        2,
        12,
        2,
        52,
        1,
        2,
        1,
        2,
        4,
        2,
        57,
        8,
        2,
        11,
        2,
        12,
        2,
        58,
        3,
        2,
        61,
        8,
        2,
        1,
        3,
        1,
        3,
        5,
        3,
        65,
        8,
        3,
        10,
        3,
        12,
        3,
        68,
        9,
        3,
        1,
        3,
        3,
        3,
        71,
        8,
        3,
        1,
        4,
        1,
        4,
        1,
        4,
        1,
        4,
        1,
        4,
        1,
        4,
        5,
        4,
        79,
        8,
        4,
        10,
        4,
        12,
        4,
        82,
        9,
        4,
        1,
        4,
        1,
        4,
        1,
        4,
        1,
        4,
        1,
        4,
        1,
        4,
        1,
        4,
        5,
        4,
        91,
        8,
        4,
        10,
        4,
        12,
        4,
        94,
        9,
        4,
        3,
        4,
        96,
        8,
        4,
        1,
        5,
        1,
        5,
        1,
        5,
        1,
        5,
        1,
        5,
        3,
        5,
        103,
        8,
        5,
        1,
        6,
        1,
        6,
        1,
        7,
        1,
        7,
        1,
        7,
        0,
        0,
        8,
        0,
        2,
        4,
        6,
        8,
        10,
        12,
        14,
        0,
        1,
        1,
        0,
        15,
        17,
        119,
        0,
        16,
        1,
        0,
        0,
        0,
        2,
        46,
        1,
        0,
        0,
        0,
        4,
        60,
        1,
        0,
        0,
        0,
        6,
        66,
        1,
        0,
        0,
        0,
        8,
        95,
        1,
        0,
        0,
        0,
        10,
        102,
        1,
        0,
        0,
        0,
        12,
        104,
        1,
        0,
        0,
        0,
        14,
        106,
        1,
        0,
        0,
        0,
        16,
        17,
        3,
        6,
        3,
        0,
        17,
        18,
        5,
        0,
        0,
        1,
        18,
        1,
        1,
        0,
        0,
        0,
        19,
        22,
        5,
        2,
        0,
        0,
        20,
        21,
        5,
        11,
        0,
        0,
        21,
        23,
        5,
        10,
        0,
        0,
        22,
        20,
        1,
        0,
        0,
        0,
        22,
        23,
        1,
        0,
        0,
        0,
        23,
        24,
        1,
        0,
        0,
        0,
        24,
        25,
        5,
        11,
        0,
        0,
        25,
        30,
        3,
        8,
        4,
        0,
        26,
        27,
        5,
        5,
        0,
        0,
        27,
        28,
        3,
        6,
        3,
        0,
        28,
        29,
        5,
        6,
        0,
        0,
        29,
        31,
        1,
        0,
        0,
        0,
        30,
        26,
        1,
        0,
        0,
        0,
        30,
        31,
        1,
        0,
        0,
        0,
        31,
        47,
        1,
        0,
        0,
        0,
        32,
        33,
        5,
        2,
        0,
        0,
        33,
        34,
        5,
        10,
        0,
        0,
        34,
        35,
        5,
        11,
        0,
        0,
        35,
        47,
        3,
        8,
        4,
        0,
        36,
        37,
        5,
        2,
        0,
        0,
        37,
        47,
        5,
        11,
        0,
        0,
        38,
        39,
        5,
        2,
        0,
        0,
        39,
        40,
        5,
        11,
        0,
        0,
        40,
        41,
        3,
        8,
        4,
        0,
        41,
        42,
        5,
        5,
        0,
        0,
        42,
        43,
        3,
        6,
        3,
        0,
        43,
        47,
        1,
        0,
        0,
        0,
        44,
        45,
        5,
        2,
        0,
        0,
        45,
        47,
        3,
        8,
        4,
        0,
        46,
        19,
        1,
        0,
        0,
        0,
        46,
        32,
        1,
        0,
        0,
        0,
        46,
        36,
        1,
        0,
        0,
        0,
        46,
        38,
        1,
        0,
        0,
        0,
        46,
        44,
        1,
        0,
        0,
        0,
        47,
        3,
        1,
        0,
        0,
        0,
        48,
        49,
        5,
        3,
        0,
        0,
        49,
        51,
        3,
        2,
        1,
        0,
        50,
        48,
        1,
        0,
        0,
        0,
        51,
        52,
        1,
        0,
        0,
        0,
        52,
        50,
        1,
        0,
        0,
        0,
        52,
        53,
        1,
        0,
        0,
        0,
        53,
        61,
        1,
        0,
        0,
        0,
        54,
        55,
        5,
        3,
        0,
        0,
        55,
        57,
        3,
        14,
        7,
        0,
        56,
        54,
        1,
        0,
        0,
        0,
        57,
        58,
        1,
        0,
        0,
        0,
        58,
        56,
        1,
        0,
        0,
        0,
        58,
        59,
        1,
        0,
        0,
        0,
        59,
        61,
        1,
        0,
        0,
        0,
        60,
        50,
        1,
        0,
        0,
        0,
        60,
        56,
        1,
        0,
        0,
        0,
        61,
        5,
        1,
        0,
        0,
        0,
        62,
        65,
        3,
        14,
        7,
        0,
        63,
        65,
        3,
        2,
        1,
        0,
        64,
        62,
        1,
        0,
        0,
        0,
        64,
        63,
        1,
        0,
        0,
        0,
        65,
        68,
        1,
        0,
        0,
        0,
        66,
        64,
        1,
        0,
        0,
        0,
        66,
        67,
        1,
        0,
        0,
        0,
        67,
        70,
        1,
        0,
        0,
        0,
        68,
        66,
        1,
        0,
        0,
        0,
        69,
        71,
        3,
        4,
        2,
        0,
        70,
        69,
        1,
        0,
        0,
        0,
        70,
        71,
        1,
        0,
        0,
        0,
        71,
        7,
        1,
        0,
        0,
        0,
        72,
        73,
        5,
        9,
        0,
        0,
        73,
        96,
        5,
        13,
        0,
        0,
        74,
        75,
        5,
        9,
        0,
        0,
        75,
        80,
        3,
        10,
        5,
        0,
        76,
        77,
        5,
        14,
        0,
        0,
        77,
        79,
        3,
        10,
        5,
        0,
        78,
        76,
        1,
        0,
        0,
        0,
        79,
        82,
        1,
        0,
        0,
        0,
        80,
        78,
        1,
        0,
        0,
        0,
        80,
        81,
        1,
        0,
        0,
        0,
        81,
        83,
        1,
        0,
        0,
        0,
        82,
        80,
        1,
        0,
        0,
        0,
        83,
        84,
        5,
        13,
        0,
        0,
        84,
        96,
        1,
        0,
        0,
        0,
        85,
        96,
        5,
        9,
        0,
        0,
        86,
        87,
        5,
        9,
        0,
        0,
        87,
        92,
        3,
        10,
        5,
        0,
        88,
        89,
        5,
        14,
        0,
        0,
        89,
        91,
        3,
        10,
        5,
        0,
        90,
        88,
        1,
        0,
        0,
        0,
        91,
        94,
        1,
        0,
        0,
        0,
        92,
        90,
        1,
        0,
        0,
        0,
        92,
        93,
        1,
        0,
        0,
        0,
        93,
        96,
        1,
        0,
        0,
        0,
        94,
        92,
        1,
        0,
        0,
        0,
        95,
        72,
        1,
        0,
        0,
        0,
        95,
        74,
        1,
        0,
        0,
        0,
        95,
        85,
        1,
        0,
        0,
        0,
        95,
        86,
        1,
        0,
        0,
        0,
        96,
        9,
        1,
        0,
        0,
        0,
        97,
        98,
        5,
        18,
        0,
        0,
        98,
        99,
        5,
        19,
        0,
        0,
        99,
        103,
        3,
        12,
        6,
        0,
        100,
        103,
        5,
        18,
        0,
        0,
        101,
        103,
        3,
        12,
        6,
        0,
        102,
        97,
        1,
        0,
        0,
        0,
        102,
        100,
        1,
        0,
        0,
        0,
        102,
        101,
        1,
        0,
        0,
        0,
        103,
        11,
        1,
        0,
        0,
        0,
        104,
        105,
        7,
        0,
        0,
        0,
        105,
        13,
        1,
        0,
        0,
        0,
        106,
        107,
        5,
        4,
        0,
        0,
        107,
        15,
        1,
        0,
        0,
        0,
        13,
        22,
        30,
        46,
        52,
        58,
        60,
        64,
        66,
        70,
        80,
        92,
        95,
        102,
    ]


class TagTemplateParser(Parser):

    grammarFileName = "TagTemplateParser.g4"

    atn = ATNDeserializer().deserialize(serializedATN())

    decisionsToDFA = [DFA(ds, i) for i, ds in enumerate(atn.decisionToState)]

    sharedContextCache = PredictionContextCache()

    literalNames = [
        "<INVALID>",
        "<INVALID>",
        "'%'",
        "'|'",
        "<INVALID>",
        "'{'",
        "'}'",
        "<INVALID>",
        "<INVALID>",
        "'('",
        "'.'",
        "<INVALID>",
        "<INVALID>",
        "')'",
        "','",
        "<INVALID>",
        "<INVALID>",
        "<INVALID>",
        "<INVALID>",
        "'='",
    ]

    symbolicNames = [
        "<INVALID>",
        "GLOBAL_WHITESPACE",
        "TAG_START",
        "PIPE",
        "TEXT",
        "CONTEXT_START",
        "CONTEXT_END",
        "ANY",
        "TAG_WHITESPACE",
        "ARGS_START",
        "CATEGORY_SEPARATOR",
        "TAG_ID",
        "ARGS_WHITESPACE",
        "ARGS_END",
        "ARG_SEPARATOR",
        "NUMERIC_VALUE",
        "BOOLEAN_VALUE",
        "STRING_VALUE",
        "ARG_NAME",
        "ARG_EQUALS",
    ]

    RULE_rootPattern = 0
    RULE_tag = 1
    RULE_pipeList = 2
    RULE_pattern = 3
    RULE_argumentList = 4
    RULE_argument = 5
    RULE_argumentValue = 6
    RULE_rawText = 7

    ruleNames = [
        "rootPattern",
        "tag",
        "pipeList",
        "pattern",
        "argumentList",
        "argument",
        "argumentValue",
        "rawText",
    ]

    EOF = Token.EOF
    GLOBAL_WHITESPACE = 1
    TAG_START = 2
    PIPE = 3
    TEXT = 4
    CONTEXT_START = 5
    CONTEXT_END = 6
    ANY = 7
    TAG_WHITESPACE = 8
    ARGS_START = 9
    CATEGORY_SEPARATOR = 10
    TAG_ID = 11
    ARGS_WHITESPACE = 12
    ARGS_END = 13
    ARG_SEPARATOR = 14
    NUMERIC_VALUE = 15
    BOOLEAN_VALUE = 16
    STRING_VALUE = 17
    ARG_NAME = 18
    ARG_EQUALS = 19

    def __init__(self, input: TokenStream, output: TextIO = sys.stdout):
        super().__init__(input, output)
        self.checkVersion("4.10.1")
        self._interp = ParserATNSimulator(
            self, self.atn, self.decisionsToDFA, self.sharedContextCache
        )
        self._predicates = None

    class RootPatternContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def pattern(self):
            return self.getTypedRuleContext(TagTemplateParser.PatternContext, 0)

        def EOF(self):
            return self.getToken(TagTemplateParser.EOF, 0)

        def getRuleIndex(self):
            return TagTemplateParser.RULE_rootPattern

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitRootPattern"):
                return visitor.visitRootPattern(self)
            else:
                return visitor.visitChildren(self)

    def rootPattern(self):

        localctx = TagTemplateParser.RootPatternContext(self, self._ctx, self.state)
        self.enterRule(localctx, 0, self.RULE_rootPattern)
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 16
            self.pattern()
            self.state = 17
            self.match(TagTemplateParser.EOF)
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class TagContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser
            self.categoryId = None  # Token
            self.tagId = None  # Token
            self.errorMissingCategoryId = None  # Token
            self.errorNoArgumentList = None  # Token
            self.errorUnclosedContext = None  # Token
            self.errorMissingTagId = None  # Token

        def TAG_START(self):
            return self.getToken(TagTemplateParser.TAG_START, 0)

        def argumentList(self):
            return self.getTypedRuleContext(TagTemplateParser.ArgumentListContext, 0)

        def TAG_ID(self, i: int = None):
            if i is None:
                return self.getTokens(TagTemplateParser.TAG_ID)
            else:
                return self.getToken(TagTemplateParser.TAG_ID, i)

        def CATEGORY_SEPARATOR(self):
            return self.getToken(TagTemplateParser.CATEGORY_SEPARATOR, 0)

        def CONTEXT_START(self):
            return self.getToken(TagTemplateParser.CONTEXT_START, 0)

        def pattern(self):
            return self.getTypedRuleContext(TagTemplateParser.PatternContext, 0)

        def CONTEXT_END(self):
            return self.getToken(TagTemplateParser.CONTEXT_END, 0)

        def getRuleIndex(self):
            return TagTemplateParser.RULE_tag

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitTag"):
                return visitor.visitTag(self)
            else:
                return visitor.visitChildren(self)

    def tag(self):

        localctx = TagTemplateParser.TagContext(self, self._ctx, self.state)
        self.enterRule(localctx, 2, self.RULE_tag)
        self._la = 0  # Token type
        try:
            self.state = 46
            self._errHandler.sync(self)
            la_ = self._interp.adaptivePredict(self._input, 2, self._ctx)
            if la_ == 1:
                self.enterOuterAlt(localctx, 1)
                self.state = 19
                self.match(TagTemplateParser.TAG_START)
                self.state = 22
                self._errHandler.sync(self)
                la_ = self._interp.adaptivePredict(self._input, 0, self._ctx)
                if la_ == 1:
                    self.state = 20
                    localctx.categoryId = self.match(TagTemplateParser.TAG_ID)
                    self.state = 21
                    self.match(TagTemplateParser.CATEGORY_SEPARATOR)

                self.state = 24
                localctx.tagId = self.match(TagTemplateParser.TAG_ID)
                self.state = 25
                self.argumentList()
                self.state = 30
                self._errHandler.sync(self)
                _la = self._input.LA(1)
                if _la == TagTemplateParser.CONTEXT_START:
                    self.state = 26
                    self.match(TagTemplateParser.CONTEXT_START)
                    self.state = 27
                    self.pattern()
                    self.state = 28
                    self.match(TagTemplateParser.CONTEXT_END)

                pass

            elif la_ == 2:
                self.enterOuterAlt(localctx, 2)
                self.state = 32
                self.match(TagTemplateParser.TAG_START)
                self.state = 33
                localctx.errorMissingCategoryId = self.match(
                    TagTemplateParser.CATEGORY_SEPARATOR
                )
                self.state = 34
                localctx.tagId = self.match(TagTemplateParser.TAG_ID)
                self.state = 35
                self.argumentList()
                pass

            elif la_ == 3:
                self.enterOuterAlt(localctx, 3)
                self.state = 36
                self.match(TagTemplateParser.TAG_START)
                self.state = 37
                localctx.errorNoArgumentList = self.match(TagTemplateParser.TAG_ID)
                pass

            elif la_ == 4:
                self.enterOuterAlt(localctx, 4)
                self.state = 38
                self.match(TagTemplateParser.TAG_START)
                self.state = 39
                localctx.tagId = self.match(TagTemplateParser.TAG_ID)
                self.state = 40
                self.argumentList()
                self.state = 41
                localctx.errorUnclosedContext = self.match(
                    TagTemplateParser.CONTEXT_START
                )
                self.state = 42
                self.pattern()
                pass

            elif la_ == 5:
                self.enterOuterAlt(localctx, 5)
                self.state = 44
                localctx.errorMissingTagId = self.match(TagTemplateParser.TAG_START)
                self.state = 45
                self.argumentList()
                pass

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class PipeListContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser
            self.errorNonTagInPipeList = None  # RawTextContext

        def PIPE(self, i: int = None):
            if i is None:
                return self.getTokens(TagTemplateParser.PIPE)
            else:
                return self.getToken(TagTemplateParser.PIPE, i)

        def tag(self, i: int = None):
            if i is None:
                return self.getTypedRuleContexts(TagTemplateParser.TagContext)
            else:
                return self.getTypedRuleContext(TagTemplateParser.TagContext, i)

        def rawText(self, i: int = None):
            if i is None:
                return self.getTypedRuleContexts(TagTemplateParser.RawTextContext)
            else:
                return self.getTypedRuleContext(TagTemplateParser.RawTextContext, i)

        def getRuleIndex(self):
            return TagTemplateParser.RULE_pipeList

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitPipeList"):
                return visitor.visitPipeList(self)
            else:
                return visitor.visitChildren(self)

    def pipeList(self):

        localctx = TagTemplateParser.PipeListContext(self, self._ctx, self.state)
        self.enterRule(localctx, 4, self.RULE_pipeList)
        try:
            self.state = 60
            self._errHandler.sync(self)
            la_ = self._interp.adaptivePredict(self._input, 5, self._ctx)
            if la_ == 1:
                self.enterOuterAlt(localctx, 1)
                self.state = 50
                self._errHandler.sync(self)
                _alt = 1
                while _alt != 2 and _alt != ATN.INVALID_ALT_NUMBER:
                    if _alt == 1:
                        self.state = 48
                        self.match(TagTemplateParser.PIPE)
                        self.state = 49
                        self.tag()

                    else:
                        raise NoViableAltException(self)
                    self.state = 52
                    self._errHandler.sync(self)
                    _alt = self._interp.adaptivePredict(self._input, 3, self._ctx)

                pass

            elif la_ == 2:
                self.enterOuterAlt(localctx, 2)
                self.state = 56
                self._errHandler.sync(self)
                _alt = 1
                while _alt != 2 and _alt != ATN.INVALID_ALT_NUMBER:
                    if _alt == 1:
                        self.state = 54
                        self.match(TagTemplateParser.PIPE)
                        self.state = 55
                        localctx.errorNonTagInPipeList = self.rawText()

                    else:
                        raise NoViableAltException(self)
                    self.state = 58
                    self._errHandler.sync(self)
                    _alt = self._interp.adaptivePredict(self._input, 4, self._ctx)

                pass

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class PatternContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser
            self.pipe_list = None  # PipeListContext

        def rawText(self, i: int = None):
            if i is None:
                return self.getTypedRuleContexts(TagTemplateParser.RawTextContext)
            else:
                return self.getTypedRuleContext(TagTemplateParser.RawTextContext, i)

        def tag(self, i: int = None):
            if i is None:
                return self.getTypedRuleContexts(TagTemplateParser.TagContext)
            else:
                return self.getTypedRuleContext(TagTemplateParser.TagContext, i)

        def pipeList(self):
            return self.getTypedRuleContext(TagTemplateParser.PipeListContext, 0)

        def getRuleIndex(self):
            return TagTemplateParser.RULE_pattern

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitPattern"):
                return visitor.visitPattern(self)
            else:
                return visitor.visitChildren(self)

    def pattern(self):

        localctx = TagTemplateParser.PatternContext(self, self._ctx, self.state)
        self.enterRule(localctx, 6, self.RULE_pattern)
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 66
            self._errHandler.sync(self)
            _alt = self._interp.adaptivePredict(self._input, 7, self._ctx)
            while _alt != 2 and _alt != ATN.INVALID_ALT_NUMBER:
                if _alt == 1:
                    self.state = 64
                    self._errHandler.sync(self)
                    token = self._input.LA(1)
                    if token in [TagTemplateParser.TEXT]:
                        self.state = 62
                        self.rawText()
                        pass
                    elif token in [TagTemplateParser.TAG_START]:
                        self.state = 63
                        self.tag()
                        pass
                    else:
                        raise NoViableAltException(self)

                self.state = 68
                self._errHandler.sync(self)
                _alt = self._interp.adaptivePredict(self._input, 7, self._ctx)

            self.state = 70
            self._errHandler.sync(self)
            la_ = self._interp.adaptivePredict(self._input, 8, self._ctx)
            if la_ == 1:
                self.state = 69
                localctx.pipe_list = self.pipeList()

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class ArgumentListContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser
            self.errorUnclosedArgumentList = None  # Token

        def ARGS_START(self):
            return self.getToken(TagTemplateParser.ARGS_START, 0)

        def ARGS_END(self):
            return self.getToken(TagTemplateParser.ARGS_END, 0)

        def argument(self, i: int = None):
            if i is None:
                return self.getTypedRuleContexts(TagTemplateParser.ArgumentContext)
            else:
                return self.getTypedRuleContext(TagTemplateParser.ArgumentContext, i)

        def ARG_SEPARATOR(self, i: int = None):
            if i is None:
                return self.getTokens(TagTemplateParser.ARG_SEPARATOR)
            else:
                return self.getToken(TagTemplateParser.ARG_SEPARATOR, i)

        def getRuleIndex(self):
            return TagTemplateParser.RULE_argumentList

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitArgumentList"):
                return visitor.visitArgumentList(self)
            else:
                return visitor.visitChildren(self)

    def argumentList(self):

        localctx = TagTemplateParser.ArgumentListContext(self, self._ctx, self.state)
        self.enterRule(localctx, 8, self.RULE_argumentList)
        self._la = 0  # Token type
        try:
            self.state = 95
            self._errHandler.sync(self)
            la_ = self._interp.adaptivePredict(self._input, 11, self._ctx)
            if la_ == 1:
                self.enterOuterAlt(localctx, 1)
                self.state = 72
                self.match(TagTemplateParser.ARGS_START)
                self.state = 73
                self.match(TagTemplateParser.ARGS_END)
                pass

            elif la_ == 2:
                self.enterOuterAlt(localctx, 2)
                self.state = 74
                self.match(TagTemplateParser.ARGS_START)
                self.state = 75
                self.argument()
                self.state = 80
                self._errHandler.sync(self)
                _la = self._input.LA(1)
                while _la == TagTemplateParser.ARG_SEPARATOR:
                    self.state = 76
                    self.match(TagTemplateParser.ARG_SEPARATOR)
                    self.state = 77
                    self.argument()
                    self.state = 82
                    self._errHandler.sync(self)
                    _la = self._input.LA(1)

                self.state = 83
                self.match(TagTemplateParser.ARGS_END)
                pass

            elif la_ == 3:
                self.enterOuterAlt(localctx, 3)
                self.state = 85
                localctx.errorUnclosedArgumentList = self.match(
                    TagTemplateParser.ARGS_START
                )
                pass

            elif la_ == 4:
                self.enterOuterAlt(localctx, 4)
                self.state = 86
                localctx.errorUnclosedArgumentList = self.match(
                    TagTemplateParser.ARGS_START
                )
                self.state = 87
                self.argument()
                self.state = 92
                self._errHandler.sync(self)
                _la = self._input.LA(1)
                while _la == TagTemplateParser.ARG_SEPARATOR:
                    self.state = 88
                    self.match(TagTemplateParser.ARG_SEPARATOR)
                    self.state = 89
                    self.argument()
                    self.state = 94
                    self._errHandler.sync(self)
                    _la = self._input.LA(1)

                pass

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class ArgumentContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def ARG_NAME(self):
            return self.getToken(TagTemplateParser.ARG_NAME, 0)

        def ARG_EQUALS(self):
            return self.getToken(TagTemplateParser.ARG_EQUALS, 0)

        def argumentValue(self):
            return self.getTypedRuleContext(TagTemplateParser.ArgumentValueContext, 0)

        def getRuleIndex(self):
            return TagTemplateParser.RULE_argument

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitArgument"):
                return visitor.visitArgument(self)
            else:
                return visitor.visitChildren(self)

    def argument(self):

        localctx = TagTemplateParser.ArgumentContext(self, self._ctx, self.state)
        self.enterRule(localctx, 10, self.RULE_argument)
        try:
            self.state = 102
            self._errHandler.sync(self)
            la_ = self._interp.adaptivePredict(self._input, 12, self._ctx)
            if la_ == 1:
                self.enterOuterAlt(localctx, 1)
                self.state = 97
                self.match(TagTemplateParser.ARG_NAME)
                self.state = 98
                self.match(TagTemplateParser.ARG_EQUALS)
                self.state = 99
                self.argumentValue()
                pass

            elif la_ == 2:
                self.enterOuterAlt(localctx, 2)
                self.state = 100
                self.match(TagTemplateParser.ARG_NAME)
                pass

            elif la_ == 3:
                self.enterOuterAlt(localctx, 3)
                self.state = 101
                self.argumentValue()
                pass

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class ArgumentValueContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def BOOLEAN_VALUE(self):
            return self.getToken(TagTemplateParser.BOOLEAN_VALUE, 0)

        def NUMERIC_VALUE(self):
            return self.getToken(TagTemplateParser.NUMERIC_VALUE, 0)

        def STRING_VALUE(self):
            return self.getToken(TagTemplateParser.STRING_VALUE, 0)

        def getRuleIndex(self):
            return TagTemplateParser.RULE_argumentValue

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitArgumentValue"):
                return visitor.visitArgumentValue(self)
            else:
                return visitor.visitChildren(self)

    def argumentValue(self):

        localctx = TagTemplateParser.ArgumentValueContext(self, self._ctx, self.state)
        self.enterRule(localctx, 12, self.RULE_argumentValue)
        self._la = 0  # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 104
            _la = self._input.LA(1)
            if not (
                (
                    ((_la) & ~0x3F) == 0
                    and (
                        (1 << _la)
                        & (
                            (1 << TagTemplateParser.NUMERIC_VALUE)
                            | (1 << TagTemplateParser.BOOLEAN_VALUE)
                            | (1 << TagTemplateParser.STRING_VALUE)
                        )
                    )
                    != 0
                )
            ):
                self._errHandler.recoverInline(self)
            else:
                self._errHandler.reportMatch(self)
                self.consume()
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class RawTextContext(ParserRuleContext):
        __slots__ = "parser"

        def __init__(
            self, parser, parent: ParserRuleContext = None, invokingState: int = -1
        ):
            super().__init__(parent, invokingState)
            self.parser = parser

        def TEXT(self):
            return self.getToken(TagTemplateParser.TEXT, 0)

        def getRuleIndex(self):
            return TagTemplateParser.RULE_rawText

        def accept(self, visitor: ParseTreeVisitor):
            if hasattr(visitor, "visitRawText"):
                return visitor.visitRawText(self)
            else:
                return visitor.visitChildren(self)

    def rawText(self):

        localctx = TagTemplateParser.RawTextContext(self, self._ctx, self.state)
        self.enterRule(localctx, 14, self.RULE_rawText)
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 106
            self.match(TagTemplateParser.TEXT)
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx



import datetime
import itertools
import mimetypes
from collections import defaultdict
from math import ceil, floor
from pathlib import Path
from typing import Any, Optional, Union

import isodate
import magic
import pathvalidate

from tempren.path_generator import evaluate_expression
from tempren.template.path_generators import File
from tempren.template.tree_elements import Tag

mimetypes.init()


class CountTag(Tag):
    """Generates sequential numbers for each invocation"""

    require_context = False
    step: int
    width: int
    _common_counter: Optional[int] = None
    _per_directory_counters: defaultdict

    def configure(self, start: int = 0, step: int = 1, width: int = 0, common: bool = False):  # type: ignore
        """
        :param start: first element of generated sequence
        :param step: step added to previous element on each invocation
        :param width: number of characters taken by the number (will be zero-filled)
        :param common: use a single, global counter instead of per-directory ones

        If width is greater than zero, output will be a string value.
        """
        if start < 0:
            raise ValueError("start have to be greater or equal 0")
        if step == 0:
            raise ValueError("step cannot be equal 0")
        self.step = step
        if width < 0:
            raise ValueError("width have to be greater or equal 0")
        self.width = width
        if common:
            self._common_counter = start
        self._per_directory_counters = defaultdict(lambda: start)

    def process(self, file: File, context: Optional[str]) -> Union[str, int]:
        value = self._get_counter_value_for(file)
        if self.width != 0:
            return str(value).zfill(self.width)
        return value

    def _get_counter_value_for(self, file: File) -> int:
        directory = file.absolute_path.parent
        if self._common_counter is not None:
            value = self._common_counter
            self._common_counter += self.step
        else:
            value = self._per_directory_counters[directory]
            self._per_directory_counters[directory] += self.step
        if value < 0:
            raise ValueError(f"Invalid counter value generated: {value}")
        return value


class ExtTag(Tag):
    """File extension (with leading dot)

    If no context is provided, current file path is used and extension is extracted from it.
    If context is present, it is parsed as a path and file extension is extracted from it.
    """

    require_context = None

    def process(self, file: File, context: Optional[str]) -> str:
        if context:
            return str(Path(context).suffix)
        return str(file.relative_path.suffix)


class BaseTag(Tag):
    """Base file name without extension (suffix)

    If no context is provided, current file path is used to determine the base file name.
    If context is present, it is parsed as a path and file name is extracted from it.
    """

    require_context = None

    def process(self, file: File, context: Optional[str]) -> str:
        if context:
            return Path(context).stem
        return file.relative_path.stem


class DirTag(Tag):
    """Directory containing processed file

    If no context is provided, current file path is used to determine the parent directory.
    If context is present, it is parsed as a path to extract the parent directory.
    """

    require_context = None

    def process(self, file: File, context: Optional[str]) -> Path:
        if context:
            return Path(context).parent
        return file.relative_path.parent


class NameTag(Tag):
    """File name (basename with extension)"""

    require_context = None

    def process(self, file: File, context: Optional[str]) -> str:
        if context:
            return str(Path(context).name)
        return str(file.relative_path.name)


class SanitizeTag(Tag):
    """Removes filesystem-unsafe characters from provided context"""

    require_context = True

    def process(self, file: File, context: Optional[str]) -> str:
        assert context
        return str(pathvalidate.sanitize_filepath(context))


class MimeTag(Tag):
    """MIME type of processed file"""

    require_context = False
    select_type: bool = False
    select_subtype: bool = False

    def configure(self, type: bool = False, subtype: bool = False):  # type: ignore
        """
        :param type: return just type section of the mime type
        :param subtype: return just subtype section of the mime type
        """
        self.select_type = type
        self.select_subtype = subtype

    def process(self, file: File, context: Optional[str]) -> str:
        mime_type = magic.from_file(file.absolute_path, mime=True)
        if self.select_type and not self.select_subtype:
            return mime_type.split("/")[0]
        elif not self.select_type and self.select_subtype:
            return mime_type.split("/")[1]
        return mime_type


class MimeExtTag(Tag):
    """File extension guessed from MIME type"""

    require_context = False

    def process(self, file: File, context: Optional[str]) -> str:
        mime_type = magic.from_file(file.absolute_path, mime=True)
        return str(mimetypes.guess_extension(mime_type, False))


class IsMimeTag(Tag):
    """Checks if processed file MIME type matches provided value"""

    require_context = False
    expected_type_prefix: str

    def configure(self, type_prefix: str):  # type: ignore
        """
        :param type_prefix: expected MIME type
        :returns: True if file MIME type starts with expected value, False otherwise
        """
        self.expected_type_prefix = type_prefix

    def process(self, file: File, context: Optional[str]) -> Any:
        mime_type = magic.from_file(file.absolute_path, mime=True)
        return mime_type.startswith(self.expected_type_prefix)


class DefaultTag(Tag):
    """Returns default value if context is empty"""

    require_context = True

    default_value: Any

    def configure(self, default_value):  # type: ignore
        """
        :param default_value: value to be used if the context is empty or consists only of whitespace
        """
        self.default_value = default_value

    def process(self, file: File, context: Optional[str]) -> Any:
        if context and not context.isspace():
            return context
        return self.default_value


class AsSizeTag(Tag):
    """Change size (in bytes) representation

    Supported unit prefixes are: K, M, G, T and P
    """

    require_context = True

    target_unit_multiplier: int
    precision_digits: Optional[int]

    _unit_weights = {
        "k": 1024,
        "m": 1024 ** 2,
        "g": 1024 ** 3,
        "t": 1024 ** 4,
        "p": 1024 ** 5,
    }

    def configure(self, unit: str, ndigits: Optional[int] = None):  # type: ignore
        """
        :param unit: name of the target unit
        :param ndigits: number of decimal digits
        """
        unit = str(unit).lower()
        unit_prefix = next(
            filter(
                lambda prefix: unit.startswith(prefix), AsSizeTag._unit_weights.keys()
            ),
            None,
        )
        if unit_prefix is None:
            raise ValueError("Unknown unit provided")
        self.target_unit_multiplier = AsSizeTag._unit_weights[unit_prefix]
        if ndigits is not None and ndigits <= 0:
            raise ValueError("Precision have to be positive")
        self.precision_digits = ndigits

    def process(self, file: File, context: Optional[str]) -> str:
        assert context is not None
        size_in_bytes = float(context)
        size_in_target_unit = size_in_bytes / self.target_unit_multiplier
        size_in_target_unit = round(size_in_target_unit, self.precision_digits)
        return str(size_in_target_unit)


class RoundTag(Tag):
    """Round provided numeric value to specified number of decimal digits"""

    require_context = True

    precision_digits: int
    direction: Optional[bool]

    def configure(self, ndigits: int = 0, down: bool = False, up: bool = False):  # type: ignore
        """
        :param ndigits: number of decimal digits
        :param down: round to the closest integer less than the processed value
        :param up: round to the closest integer greater than the processed value
        """
        if ndigits != 0 and (down or up):
            raise ValueError("down or up cannot be used with precision digits")
        if down and up:
            raise ValueError("down and up cannot cannot be specified simultaneously")
        if down:
            self.direction = False
        elif up:
            self.direction = True
        else:
            self.direction = None
        self.precision_digits = ndigits

    def process(self, file: File, context: Optional[str]) -> Any:
        assert context is not None
        number = float(context)
        if self.direction is None:
            rounded_number = round(number, self.precision_digits)
            if self.precision_digits <= 0:
                rounded_number = int(rounded_number)
        elif self.direction:
            rounded_number = ceil(number)
        else:
            rounded_number = floor(number)
        return str(rounded_number)


class EvalTag(Tag):
    """Evaluate context as a Python expression"""

    require_context = True

    def configure(self):
        super().configure()

    def process(self, file: File, context: Optional[str]) -> Any:  # type ignore
        assert context is not None
        return evaluate_expression(context)


class AsTimeTag(Tag):
    """Re-format provided date-time (in ISO 8601 representation)"""

    require_context = True

    destination_format: str

    def configure(self, format: str):  # type: ignore
        """
        :param format: strftime format to use
        """
        self.destination_format = format

    def process(self, file: File, context: Optional[str]) -> str:  # type: ignore
        assert context is not None
        parsed_datetime = datetime.datetime.fromisoformat(context)
        return parsed_datetime.strftime(self.destination_format)


AsTimeTag.__doc__ = "\n".join(
    [
        str(AsTimeTag.__doc__),
        """Possible fields:
Date
  %a\t Abbreviated weekday name ("Mon")
  %A\t Weekday name ("Monday")
  %b\t Abbreviated month name ("Oct")
  %B\t Month name ("October")
  %d\t Zero-prefixed day of the month (01..31)
  %e\t Day of the month (1..31)
  %j\t Zero-prefixed day of the year (001..366)
  %m\t Zero-prefixed month of the year (01..12)
  %U\t Week number of the current year (with Sunday being first day of the week) (00..53)
  %W\t Week number of the current year (with Monday being first day of the week) (00..53)
  %w\t Day of the week (where Sunday is 0) (0..6)
  %x\t System dependent representation for the date alone
  %y\t Year without a century (00..99)
  %Y\t Year with a century

Time
  %H\t Zero-prefixed hour (00..23)
  %I\t Zero-prefixed hour (01..12)
  %l\t Hour (1..12)
  %M\t Zero-prefixed minutes (00..59)
  %N\t Fractional seconds digits, default is 9 digits (nanosecond)
  %P\t Meridian indicator (am/pm)
  %p\t Meridian indicator (AM/PM)
  %S\t Zero-prefixed Seconds (00..60)
  %X\t System dependent representation for the time alone
  %Z\t Time zone name

Other
  %c\t System dependent date and time representation
  %%\t "%" character itself""",
    ]
)


class AsDurationTag(Tag):
    """Re-format provided duration (in ISO 8601 representation)"""

    require_context = True

    destination_format: str

    def configure(self, format: str):  # type: ignore
        """
        :param format: strftime format to use
        """
        self.destination_format = format

    def process(self, file: File, context: Optional[str]) -> str:  # type: ignore
        assert context is not None
        parsed_duration = isodate.parse_duration(context)
        return isodate.strftime(parsed_duration, self.destination_format)


AsDurationTag.__doc__ = "\n".join(
    [
        str(AsDurationTag.__doc__),
        """Possible fields:
  %j\t Zero-prefixed day of the year (001..366)
  %d\t Zero-prefixed day of the month (01..31)
  %m\t Zero-prefixed month (01..12)
  %W\t Week number of the current year (with Monday being first day of the week) (00..53)
  %w\t Day of the week (where Monday is 0) (0..6)
  %Y\t Year with a century (four digits)
  %C\t Century (00..99)
  %H\t Zero-prefixed hour (00..23)
  %M\t Zero-prefixed minutes (00..59)
  %S\t Zero-prefixed seconds (00..61)
  %f\t Zero-prefixed microsecond (0..999999)
  %z\t UTC offset in the form +HHMM or -HHMM (if present)
  %Z\t Time zone name (if present)
  %P\t ISO8601 duration format
  %p\t ISO8601 duration format in weeks
  %%\t "%" character itself""",
    ]
)


_supported_bases = {2: "b", 8: "o", 10: "d", 16: "x"}


class AsIntTag(Tag):
    """Converts integer from the context from one base to another (removing leading zeros if necessary)"""

    require_context = True

    src_base: int
    dst_base: int

    def configure(self, src_base: int = 10, dst_base: int = 10):  # type: ignore
        """
        :param src_base: source base
        :param dst_base: destination base
        Supported bases are: 2, 8, 10 and 16
        """
        if src_base not in _supported_bases:
            raise ValueError(f"Unsupported source base: {src_base}")
        if dst_base not in _supported_bases:
            raise ValueError(f"Unsupported destination base: {dst_base}")
        self.src_base = src_base
        self.dst_base = dst_base

    def process(self, file: File, context: Optional[str]) -> str:
        assert context is not None

        parsed_number = int(context, base=self.src_base)
        return format(parsed_number, _supported_bases[self.dst_base])



import logging
import os
import shutil
from abc import ABC, abstractmethod
from pathlib import Path
from typing import Callable, Iterable, Set

from tempren.path_generator import File

FileRenamerType = Callable[[Path, Path, bool], None]


class InvalidDestinationError(Exception):
    pass


class DestinationAlreadyExistsError(FileExistsError):
    def __init__(self, src: Path, dst: Path):
        super().__init__(
            f"Could not rename '{src}' as destination path '{dst}' already exists"
        )


class FileGatherer(ABC):
    include_hidden: bool = False
    """Include hidden files and directories when making the search"""

    @abstractmethod
    def gather_in(self, start_directory: Path) -> Iterable[File]:
        raise NotImplementedError()


class FilesystemGatherer(FileGatherer, ABC):
    def _include_path_in_result(self, path: Path) -> bool:
        if not self.include_hidden:
            return not path.name.startswith(".")
        return True


class FlatFileGatherer(FilesystemGatherer):
    def gather_in(self, start_directory: Path) -> Iterable[File]:
        yield from map(
            lambda file_path: File(
                start_directory, file_path.relative_to(start_directory)
            ),
            filter(
                lambda file_path: self._include_path_in_result(file_path)
                and not file_path.is_dir(),
                start_directory.glob("*"),
            ),
        )


class RecursiveFileGatherer(FilesystemGatherer):
    def gather_in(self, start_directory: Path) -> Iterable[File]:
        yield from self._gather_in(start_directory, start_directory)

    def _gather_in(self, directory: Path, start_directory: Path) -> Iterable[File]:
        for dir_entry in directory.iterdir():
            if not self._include_path_in_result(dir_entry):
                continue

            if dir_entry.is_dir():
                yield from self._gather_in(dir_entry, start_directory)
            else:
                yield File(start_directory, dir_entry.relative_to(start_directory))


class FileRenamer:
    def __call__(
        self,
        source_path: Path,
        destination_path: Path,
        override: bool = False,
    ) -> None:
        if not override and destination_path.exists():
            raise DestinationAlreadyExistsError(source_path, destination_path)
        if source_path.parent != destination_path.parent:
            raise InvalidDestinationError(
                f"Destination path {destination_path} targets different directory"
            )
        os.rename(source_path, destination_path)


class FileMover:
    def __call__(
        self,
        source_path: Path,
        destination_path: Path,
        override: bool = False,
    ) -> None:
        if not override and destination_path.exists():
            raise DestinationAlreadyExistsError(source_path, destination_path)
        destination_path.parent.mkdir(parents=True, exist_ok=True)
        shutil.move(str(source_path), destination_path)


class DryRunRenamer:
    def __init__(self):
        self.removed_paths: Set[Path] = set()
        self.created_paths: Set[Path] = set()

    def __call__(
        self,
        source_path: Path,
        destination_path: Path,
        override: bool = False,
    ) -> None:
        source_exists = (
            source_path.exists() or source_path in self.created_paths
        ) and source_path not in self.removed_paths
        if not source_exists:
            raise FileNotFoundError(f"No such file or directory: {source_path}")

        destination_exists = (
            destination_path.exists() or destination_path in self.created_paths
        ) and destination_path not in self.removed_paths
        if destination_exists and not override:
            raise FileExistsError(
                f"Destination file already exists: {destination_path}"
            )

        self.removed_paths.add(source_path)
        self.created_paths.add(destination_path)
        self.removed_paths.discard(destination_path)
        self.created_paths.discard(source_path)


class PrintingRenamerWrapper:
    log: logging.Logger

    def __init__(self, renamer_to_wrap: FileRenamerType):
        assert renamer_to_wrap is not None
        self.log = logging.getLogger(self.__class__.__name__)
        self.renamer = renamer_to_wrap

    def __call__(
        self,
        source_path: Path,
        destination_path: Path,
        override: bool = False,
    ) -> None:
        if override:
            self.log.debug(
                "Trying to override: %s to %s", source_path, destination_path
            )
        else:
            self.log.debug("Trying to rename: %s to %s", source_path, destination_path)
        self.renamer(source_path, destination_path, override)

        self.log.info(f"Renamed: {source_path}")
        if override:
            self.log.info(f"         to: {destination_path} (override)")
            # self.log.info(f"to existing: {destination_path} (override)")
        else:
            self.log.info(f"     to: {destination_path}")



import logging
import os
from dataclasses import dataclass
from enum import Enum
from pathlib import Path
from typing import Callable, Iterable, Optional, Union

from tempren.file_filters import (
    FileFilterInverter,
    GlobFilenameFileFilter,
    GlobPathFileFilter,
    RegexFilenameFileFilter,
    RegexPathFileFilter,
    TemplateFileFilter,
)
from tempren.file_sorters import TemplateFileSorter
from tempren.filesystem import (
    DestinationAlreadyExistsError,
    DryRunRenamer,
    FileGatherer,
    FileMover,
    FileRenamer,
    FileRenamerType,
    FlatFileGatherer,
    InvalidDestinationError,
    PrintingRenamerWrapper,
    RecursiveFileGatherer,
)
from tempren.path_generator import File, InvalidFilenameError, PathGenerator
from tempren.template.path_generators import (
    TemplateNameGenerator,
    TemplatePathGenerator,
)
from tempren.template.tree_builder import TagRegistry, TagTreeBuilder, TemplateError
from tempren.template.tree_elements import Pattern

log = logging.getLogger(__name__)


class FilterType(Enum):
    template = "template"
    regex = "regex"
    glob = "glob"


class OperationMode(Enum):
    name = "name"
    path = "path"


# TODO: Find a way to keep documentation close to the enum values and use it in argparser/generated help
class ConflictResolutionStrategy(Enum):
    stop = "stop"
    """Stop renaming and show an error"""

    ignore = "ignore"
    """Leave conflicting record unchanged and continue with renaming"""

    override = "override"
    """Override destination record with a new name"""

    manual = "manual"
    """Prompt user to resolve conflict manually (choose an option or provide new filename)"""


ManualConflictResolver = Callable[[Path, Path], Union[ConflictResolutionStrategy, Path]]


@dataclass
class RuntimeConfiguration:
    template: str
    input_directory: Path
    recursive: bool = False
    include_hidden: bool = False
    dry_run: bool = False
    filter_type: FilterType = FilterType.glob
    filter_invert: bool = False
    filter: Optional[str] = None
    conflict_strategy: ConflictResolutionStrategy = ConflictResolutionStrategy.stop
    sort_invert: bool = False
    sort: Optional[str] = None
    mode: OperationMode = OperationMode.name


class ConfigurationError(Exception):
    pass


def manual_resolver_placeholder(
    source_path: Path, destination_path: Path
) -> Union[ConflictResolutionStrategy, Path]:
    raise NotImplementedError()


class Pipeline:
    log: logging.Logger
    _input_directory: Path
    file_gatherer: FileGatherer
    sorter: Optional[Callable[[Iterable[File]], Iterable[File]]] = None
    path_generator: PathGenerator
    conflict_strategy: ConflictResolutionStrategy = ConflictResolutionStrategy.stop

    def __init__(self):
        self.log = logging.getLogger(__name__)
        self.file_filter: Callable[[File], bool] = lambda file: True
        self.renamer: FileRenamerType = DryRunRenamer()
        self.manual_conflict_resolver: ManualConflictResolver = (
            manual_resolver_placeholder
        )

    @property
    def input_directory(self) -> Path:
        return self._input_directory

    @input_directory.setter
    def input_directory(self, input_path: Path):
        self._input_directory = input_path.absolute()

    def execute(self):
        all_files = []
        self.log.info(f"Gathering paths in {self.input_directory}")
        os.chdir(self.input_directory)
        for file in self.file_gatherer.gather_in(self.input_directory):
            self.log.debug("Checking %s", file)
            if not self.file_filter(file):
                self.log.debug("%s filtered out", file)
                continue
            self.log.debug("%s considered for renaming", file)
            all_files.append(file)

        self.log.info("%d files considered for renaming", len(all_files))
        if self.sorter:
            self.log.info("Sorting files")
            all_files = self.sorter(all_files)

        self.log.debug("Generating new names")
        # In case when destination file exists in the first run, we add such name to the
        # backlog and try again (in reverse order) later. This should mitigate most of
        # transitional conflicts.
        backlog = []
        for file in all_files:
            try:
                self.log.debug("Generating new name for %r", file)
                new_relative_path = self.path_generator.generate(file)
                self.log.debug("Generated path: '%s'", new_relative_path)
            except InvalidFilenameError as error:
                # TODO: Introduce flag similar to conflict resolver to take appropriate action
                self.log.warning(
                    "Invalid name generated for %r: %r",
                    file,
                    error.generated_name,
                )
                raise InvalidDestinationError()
            except Exception as error:
                # TODO: Test
                self.log.error(
                    "Error generated when renaming %r: %r",
                    file,
                    error,
                )
                raise

            if new_relative_path == file.relative_path:
                self.log.info(
                    "Skipping renaming of: '%s' (source and destination are the same)",
                    new_relative_path,
                )
                continue

            new_absolute_path = (file.input_directory / new_relative_path).resolve()
            if not str(new_absolute_path).startswith(str(file.input_directory)):
                raise InvalidDestinationError(
                    f"Path generated for {file!r}: {new_relative_path} is not relative to the input directory",
                )

            try:
                self.renamer(file.relative_path, new_relative_path)
            except FileExistsError:
                self.log.debug(
                    "Deferring renaming of %r as destination '%s' already exists",
                    file,
                    new_relative_path,
                )
                backlog.append((file.relative_path, new_relative_path))

        while backlog:
            source_path, destination_path = backlog.pop()
            self.log.debug(
                "Trying again to rename '%s' into '%s'", source_path, destination_path
            )
            try:
                self.renamer(source_path, destination_path)
            except FileExistsError:
                self.resolve_conflict(
                    source_path, destination_path, self.conflict_strategy
                )

    def resolve_conflict(
        self,
        source_path: Path,
        destination_path: Path,
        strategy: ConflictResolutionStrategy,
    ):
        if strategy == ConflictResolutionStrategy.stop:
            raise DestinationAlreadyExistsError(source_path, destination_path)
        elif strategy == ConflictResolutionStrategy.ignore:
            self.log.debug(
                "Skipping renaming of '%s' to '%s' as destination path already exists",
                source_path,
                destination_path,
            )
        elif strategy == ConflictResolutionStrategy.override:
            self.log.warning(
                "Overriding destination '%s' as it already exists", destination_path
            )
            self.renamer(source_path, destination_path, True)
        elif strategy == ConflictResolutionStrategy.manual:
            if self.manual_conflict_resolver is None:
                raise NotImplementedError("Manual conflict resolver not configured")
            user_selected_strategy = self.manual_conflict_resolver(
                source_path, destination_path
            )
            if isinstance(user_selected_strategy, Path):
                user_provided_path: Path = user_selected_strategy
                self.renamer(source_path, user_provided_path, False)
            else:
                self.resolve_conflict(
                    source_path, destination_path, user_selected_strategy
                )
        else:
            raise NotImplementedError(
                f"Unknown conflict resolution strategy: {strategy}"
            )


def build_tag_registry() -> TagRegistry:
    log.debug("Building tag registry")
    import tempren.tags

    registry = TagRegistry()
    registry.register_tags_in_package(tempren.tags)
    return registry


def build_pipeline(
    config: RuntimeConfiguration,
    registry: TagRegistry,
    manual_conflict_resolver: ManualConflictResolver,  # TODO: Move to the RuntimeConfiguration
) -> Pipeline:
    log.debug("Building pipeline")
    pipeline = Pipeline()
    pipeline.input_directory = config.input_directory
    tree_builder = TagTreeBuilder()

    if config.recursive:
        pipeline.file_gatherer = RecursiveFileGatherer()
    else:
        pipeline.file_gatherer = FlatFileGatherer()

    pipeline.file_gatherer.include_hidden = config.include_hidden

    def _compile_template(template_text: str) -> Pattern:
        log.debug("Compiling template %r", template_text)
        try:
            return registry.bind(tree_builder.parse(template_text))
        except TemplateError as template_error:
            template_error.template = template_text
            raise template_error

    bound_pattern = _compile_template(config.template)

    if config.mode == OperationMode.name:
        pipeline.path_generator = TemplateNameGenerator(bound_pattern)
        if config.filter:
            if config.filter_type == FilterType.regex:
                pipeline.file_filter = RegexFilenameFileFilter(config.filter)
            elif config.filter_type == FilterType.glob:
                pipeline.file_filter = GlobFilenameFileFilter(config.filter)
            elif config.filter_type == FilterType.template:
                bound_filter_pattern = _compile_template(config.filter)
                pipeline.file_filter = TemplateFileFilter(bound_filter_pattern)
            else:
                raise NotImplementedError("Unknown filter type")
    elif config.mode == OperationMode.path:
        pipeline.path_generator = TemplatePathGenerator(bound_pattern)
        if config.filter:
            if config.filter_type == FilterType.regex:
                pipeline.file_filter = RegexPathFileFilter(config.filter)
            elif config.filter_type == FilterType.glob:
                pipeline.file_filter = GlobPathFileFilter(config.filter)
            elif config.filter_type == FilterType.template:
                bound_filter_pattern = _compile_template(config.filter)
                pipeline.file_filter = TemplateFileFilter(bound_filter_pattern)
            else:
                raise NotImplementedError("Unknown filter type")
    else:
        raise NotImplementedError("Unknown operation mode")

    if config.filter_invert and config.filter is not None:
        pipeline.file_filter = FileFilterInverter(pipeline.file_filter)

    if config.sort:
        bound_sorter_pattern = _compile_template(config.sort)
        pipeline.sorter = TemplateFileSorter(bound_sorter_pattern, config.sort_invert)

    pipeline.conflict_strategy = config.conflict_strategy
    pipeline.manual_conflict_resolver = manual_conflict_resolver

    if config.dry_run:
        pipeline.renamer = DryRunRenamer()
        # FIXME: Use DryRunMover for path mode?
    else:
        if config.mode == OperationMode.name:
            pipeline.renamer = FileRenamer()
        elif config.mode == OperationMode.path:
            pipeline.renamer = FileMover()
        else:
            raise NotImplementedError("Unknown operation mode")

    pipeline.renamer = PrintingRenamerWrapper(pipeline.renamer)  # type: ignore
    return pipeline



import textwrap
from abc import ABC, abstractmethod
from dataclasses import dataclass, field
from typing import Any, List, Mapping, Optional, Type

from docstring_parser import parse as parse_docstring

from tempren.path_generator import File


@dataclass
class TagName:
    name: str
    category: Optional[str] = None

    def __post_init__(self):
        if len(self.name) < 1:
            raise ValueError(f"Invalid tag name: ${repr(self.name)}")
        if self.category is not None and len(self.category) < 1:
            raise ValueError(f"Invalid category name: ${repr(self.category)}")

    def __str__(self) -> str:
        if self.category:
            return f"{self.category}.{self.name}"
        return f"{self.name}"


@dataclass
class Location:
    line: int
    column: int
    length: int

    def __str__(self) -> str:
        if self.length == 1:
            return f"line {self.line}:{self.column}"
        return f"line {self.line}:{self.column}-{self.column + self.length}"


class PatternElement(ABC):
    """Represents element of template pattern"""

    @abstractmethod
    def process(self, file: File) -> Any:
        raise NotImplementedError()


@dataclass
class RawText(PatternElement):
    """Represents constant text (non-tag) part of the template"""

    location: Location = field(init=False, compare=False)
    text: str

    def process(self, file: File) -> str:
        return self.text


@dataclass
class Pattern(PatternElement):
    """Represents pattern tree - a chain of text/tag invocations"""

    sub_elements: List[PatternElement] = field(default_factory=list)
    source_representation: Optional[str] = field(
        init=False, default=None, compare=False
    )

    def process(self, file: File) -> str:
        """Recursively renders pattern as a string"""
        return "".join(
            map(
                lambda se: str(se.process(file)),
                self.sub_elements,
            )
        )

    def process_as_expression(self, file: File) -> str:
        """Recursively renders pattern as an expression string

        In expression string, values returned by the tags are converted using `repr`
        rather than `str` to obtain valid python value representation.
        """
        return "".join(
            map(
                lambda se: Pattern._convert_to_representation(se, file),
                self.sub_elements,
            )
        )

    @staticmethod
    def _convert_to_representation(element: PatternElement, file: File) -> str:
        """Renders value returned by tag invocation as a string representation (as to be used in evaluated
        expressions)"""
        tag_value = element.process(file)
        if isinstance(element, TagInstance):
            return repr(tag_value)
        return str(tag_value)


@dataclass
class TagPlaceholder(PatternElement):
    """Represents unbound tag"""

    location: Location = field(init=False, compare=False)
    tag_name: TagName
    context: Optional[Pattern] = None
    args: List[Any] = field(default_factory=list)
    kwargs: Mapping[str, Any] = field(default_factory=dict)

    def process(self, file: File) -> str:
        raise NotImplementedError(
            "TagPlaceholder shouldn't be present in bound tag tree"
        )


class FileNotSupportedError(Exception):
    """Tag cannot extract value due to invalid file type"""

    pass


class MissingMetadataError(Exception):
    """Tag cannot extract value due to missing metadata"""

    pass


class Tag(ABC):
    require_context: Optional[bool] = None
    """Determine if tag requires context

    When set to True - an error is reported if no context is provided.
    When set to False - an error is reported if context is provided.
    When set to None - context is optional (tag decides what to do with it).
    """

    def configure(self):
        """Initialize tag instance with configuration options provided by the user"""
        pass

    @abstractmethod
    def process(self, file: File, context: Optional[str]) -> Any:
        """Execute tag logic on a single file/context"""
        raise NotImplementedError()


class TagFactory(ABC):
    @property
    @abstractmethod
    def tag_name(self) -> str:
        """Name of produced tags"""

    @property
    @abstractmethod
    def configuration_signature(self) -> str:
        """Configuration arguments"""

    @property
    @abstractmethod
    def short_description(self) -> str:
        """Short (one-line) tag description"""

    @property
    @abstractmethod
    def long_description(self) -> Optional[str]:
        """Longer tag documentation"""

    @abstractmethod
    def __call__(self, *args, **kwargs) -> Tag:
        """Creates tag from provided configuration arguments"""


class TagFactoryFromClass(TagFactory):
    _tag_class_suffix = "Tag"
    _tag_class: Type[Tag]
    _tag_name: str

    @property
    def tag_name(self) -> str:
        return self._tag_name

    @property
    def configuration_signature(self) -> str:
        single_line_signature = self._create_configuration_signature()
        argument_description = self._create_arguments_description()
        if argument_description is None:
            return single_line_signature
        return "\n".join((single_line_signature, argument_description))

    def _create_configuration_signature(self) -> str:
        import inspect

        signature = inspect.signature(self._tag_class.configure)
        parameters_without_self = list(
            filter(lambda param: param.name != "self", signature.parameters.values())
        )
        signature = signature.replace(parameters=parameters_without_self)
        if self._tag_class.require_context is None:
            context_metavar = "[{...}]"
        elif self._tag_class.require_context is True:
            context_metavar = "{...}"  # TODO: test?
        else:
            context_metavar = ""
        single_line_signature = f"%{self._tag_name}{signature}{context_metavar}"
        return single_line_signature

    def _create_arguments_description(self) -> Optional[str]:
        argument_description = []
        for argument in self._parsed_configure_docstring.params:
            argument_description.append(f"{argument.arg_name} - {argument.description}")
        if not argument_description:
            return None
        return textwrap.indent("\n".join(argument_description), "  ")

    @property
    def short_description(self) -> str:
        if self._parsed_class_docstring.short_description:
            return self._parsed_class_docstring.short_description
        return ""

    @property
    def long_description(self) -> Optional[str]:
        return self._parsed_class_docstring.long_description

    def __init__(self, tag_class: Type[Tag], tag_name: Optional[str] = None):
        self._tag_class = tag_class
        self._parsed_class_docstring = parse_docstring(
            tag_class.__doc__ if tag_class.__doc__ else ""
        )
        self._parsed_configure_docstring = parse_docstring(
            tag_class.configure.__doc__ if tag_class.configure.__doc__ else ""
        )
        if tag_name:
            self._tag_name = tag_name
        else:
            tag_class_name = tag_class.__name__
            if tag_class_name.endswith(self._tag_class_suffix):
                self._tag_name = tag_class_name[: -len(self._tag_class_suffix)]
            else:
                raise ValueError(
                    f"Could not determine tag name from tag class: {tag_class_name}"
                )

    def __call__(self, *args, **kwargs) -> Tag:
        tag = self._tag_class()
        tag.configure(*args, **kwargs)  # type: ignore
        return tag


@dataclass
class TagInstance(PatternElement):
    """Represents a tag bound to the implementation"""

    tag: Tag
    context: Optional[Pattern] = None

    def process(self, file: File) -> Any:
        context_str = self.context.process(file) if self.context else None
        try:
            return self.tag.process(file, context_str)
        except MissingMetadataError:
            return ""



import itertools
from abc import ABC, abstractmethod
from fractions import Fraction
from typing import Any, Iterator, Optional, Tuple

import piexif
import PIL
from piexif import TAGS
from piexif import TYPES as TAG_TYPES
from PIL import Image

from tempren.path_generator import File
from tempren.template.tree_elements import (
    FileNotSupportedError,
    MissingMetadataError,
    Tag,
)


class PillowTagBase(Tag, ABC):
    """Base for tags extracting metadata from images using Pillow library"""

    require_context = False

    def process(self, file: File, context: Optional[str]) -> Any:
        try:
            with Image.open(file.absolute_path) as img:
                return self.extract_metadata(img)
        except PIL.UnidentifiedImageError:
            raise FileNotSupportedError()

    @abstractmethod
    def extract_metadata(self, image: Image) -> Any:
        raise NotImplementedError()


class WidthTag(PillowTagBase):
    """Image width in pixels"""

    def extract_metadata(self, image: Image) -> Any:
        return image.width


class HeightTag(PillowTagBase):
    """Image height in pixels"""

    def extract_metadata(self, image: Image) -> Any:
        return image.height


class FormatTag(PillowTagBase):
    """Image format ('JPG', 'PNG', ...)"""

    def extract_metadata(self, image: Image) -> Any:
        return image.format


class ColorModeTag(PillowTagBase):
    """Color mode ('RGB', 'RGBA', ...)"""

    def extract_metadata(self, image: Image) -> Any:
        return image.mode


class AspectRatioTag(PillowTagBase):
    """Image aspect ratio (in fractional W:H or decimal format)"""

    use_decimal: bool

    def configure(self, decimal: bool = False):
        """
        :param decimal: use decimal notation
        """
        self.use_decimal = decimal

    def extract_metadata(self, image: Image) -> Any:
        if self.use_decimal:
            return image.width / image.height
        else:
            aspect_ratio = Fraction(image.width, image.height)
            return f"{aspect_ratio.numerator}:{aspect_ratio.denominator}"


class MPxTag(PillowTagBase):
    """Image resolution in megapixels"""

    ndigits: int

    def configure(self, ndigits: int = 2):
        """
        :param ndigits: number of decimal digits
        """
        assert ndigits >= 0, "Precision cannot be negative"
        self.ndigits = ndigits

    def extract_metadata(self, image: Image) -> Any:
        return round(image.width * image.height / 1_000_000, self.ndigits)


class IsOrientationTag(PillowTagBase):
    """Checks image orientation"""

    check_for_landscape: bool
    check_for_portrait: bool

    def configure(self, landscape: bool = False, portrait: bool = False):
        """
        :param landscape: check if image height is greater than width
        :param portrait: check if image width is greater than height
        :returns: True if image have expected orientation
        If both landscape and portrait are specified - check for squareness
        """
        # TODO: check for rotation information in EXIF data?
        assert landscape or portrait, "No orientation specified"
        self.check_for_landscape = landscape
        self.check_for_portrait = portrait

    def extract_metadata(self, image: Image) -> bool:
        if self.check_for_landscape and self.check_for_portrait:
            return image.width == image.height
        if self.check_for_landscape:
            return image.width > image.height
        if self.check_for_portrait:
            return image.width < image.height
        raise NotImplementedError()


class ExifTag(Tag):
    """Extract value of any EXIF tag"""

    tag_id: int
    tag_type: int

    def configure(self, tag_name: str):  # type: ignore
        """
        :param tag_name: name of the tag to extract (e.g. 'FocalLength', 'DateTime', 'FNumber')
        """
        # TODO: generate list of supported tags dynamically
        assert tag_name, "expected non empty tag name"
        self.tag_id, self.tag_type = self._tag_name_to_id_type(tag_name)

    @staticmethod
    def _tag_name_to_id_type(tag_name: str) -> Tuple[int, int]:
        for _, tags in TAGS.items():
            for tag_id, description in tags.items():
                if description["name"] == tag_name:
                    return tag_id, description["type"]
        raise ValueError(f"Could not find tag id for '{tag_name}'")

    def process(self, file: File, context: Optional[str]) -> Any:
        exif_dict = piexif.load(str(file.absolute_path))
        for src in exif_dict.values():
            if isinstance(src, dict) and self.tag_id in src:
                return self._extract_value(src[self.tag_id])
        else:
            raise MissingMetadataError()

    def _extract_value(self, tag_value):
        if self.tag_type in (TAG_TYPES.Rational, TAG_TYPES.SRational):
            return round(tag_value[0] / tag_value[1], 1)
        if self.tag_type == TAG_TYPES.Ascii:
            return tag_value.decode("ascii")
        return tag_value


_tag_type_map = {
    TAG_TYPES.__dict__[type_name]: type_name
    for type_name in TAG_TYPES.__dict__.keys()
    if isinstance(TAG_TYPES.__dict__[type_name], int)
}


def _generate_exif_tag_list() -> Iterator[str]:
    for _, tags in TAGS.items():
        for tag_id, description in tags.items():
            tag_name = description["name"]
            if tag_name.startswith("ZZZTest"):
                continue
            tag_type = _tag_type_map[description["type"]]
            yield f"  {tag_name} ({tag_type})"


ExifTag.__doc__ = "\n".join(
    itertools.chain(
        [str(ExifTag.__doc__), "Available tag names:"],
        sorted(set(_generate_exif_tag_list())),
    )
)



from abc import ABC
from datetime import timedelta
from typing import Any, Dict, List, Optional

import mutagen
import mutagen.mp3
from mutagen.easyid3 import EasyID3

from tempren.path_generator import File
from tempren.template.tree_elements import MissingMetadataError, Tag


class MutagenTagBase(Tag, ABC):
    """Extracts audio metadata (tags) using mutagen library"""

    require_context = False

    tag_key: str

    def process(self, file: File, context: Optional[str]) -> Any:
        metadata_dict = self._extract_metadata(file)
        if self.tag_key not in metadata_dict:
            raise MissingMetadataError()
        tag_value = self.format_value(metadata_dict[self.tag_key])
        if isinstance(tag_value, list):
            tag_value = "\n".join(tag_value)
        return tag_value

    def _extract_metadata(self, file: File) -> Dict[str, Any]:
        # TODO: Check if File(..., easy=True) could improve readability here
        audio_file = mutagen.File(file.absolute_path)
        metadata_dict = dict()
        metadata_dict["duration"] = audio_file.info.length
        metadata_dict["channels"] = audio_file.info.channels
        metadata_dict["sample_rate"] = audio_file.info.sample_rate
        metadata_dict["bitrate"] = audio_file.info.bitrate
        if isinstance(audio_file, mutagen.mp3.MP3):
            metadata_dict["comments"] = audio_file.get("COMM::XXX", "")
            audio_file = mutagen.mp3.MP3(file.absolute_path, ID3=EasyID3)
        else:
            metadata_dict["bits_per_sample"] = audio_file.info.bits_per_sample

        metadata_dict.update(dict(audio_file))
        return metadata_dict

    def format_value(self, metadata_value: Any):
        return metadata_value


class TitleTag(MutagenTagBase):
    """Track title"""

    tag_key = "title"


class AlbumTag(MutagenTagBase):
    """Album name"""

    tag_key = "album"


class ArtistTag(MutagenTagBase):
    """Name of the artist"""

    tag_key = "artist"


class CommentTag(MutagenTagBase):
    """Comment"""

    tag_key = "comments"


class YearTag(MutagenTagBase):
    """Year of the release"""

    tag_key = "date"


class GenreTag(MutagenTagBase):
    """Genre type"""

    tag_key = "genre"


class TrackTag(MutagenTagBase):
    """Track number"""

    tag_key = "tracknumber"


class DurationTag(MutagenTagBase):
    """Track duration in seconds"""

    tag_key = "duration"

    def format_value(self, metadata_value: Any):
        duration_in_seconds = float(super().format_value(metadata_value))
        return timedelta(seconds=duration_in_seconds)


class ChannelsTag(MutagenTagBase):
    """Number of channels"""

    tag_key = "channels"


class SampleRateTag(MutagenTagBase):
    """Sample rate"""

    tag_key = "sample_rate"


class BitRateTag(MutagenTagBase):
    """Bit rate"""

    tag_key = "bitrate"


class BitsPerSampleTag(MutagenTagBase):
    """Bit rate per sample"""

    tag_key = "bits_per_sample"



import hashlib
import zlib
from pathlib import Path
from typing import Optional

from tempren.path_generator import File
from tempren.template.tree_elements import Tag

CHUNK_SIZE = 4096


def _calculate_hash(algorithm, path: Path, chunk_size: int) -> str:
    with open(path, "rb") as f:
        for chunk in iter(lambda: f.read(chunk_size), b""):
            algorithm.update(chunk)
    return algorithm.hexdigest()


class Md5Tag(Tag):
    """MD5 hash of the file"""

    require_context = False

    def process(self, file: File, context: Optional[str]) -> str:
        assert context is None
        return _calculate_hash(hashlib.md5(), file.absolute_path, CHUNK_SIZE)


class Sha1Tag(Tag):
    """SHA1 hash of the file"""

    require_context = False

    def process(self, file: File, context: Optional[str]) -> str:
        assert context is None
        return _calculate_hash(hashlib.sha1(), file.absolute_path, CHUNK_SIZE)


class Sha256Tag(Tag):
    """SHA256 hash of the file"""

    require_context = False

    def process(self, file: File, context: Optional[str]) -> str:
        assert context is None
        return _calculate_hash(hashlib.sha256(), file.absolute_path, CHUNK_SIZE)


class Sha224Tag(Tag):
    """SHA224 hash of the file"""

    require_context = False

    def process(self, file: File, context: Optional[str]) -> str:
        assert context is None
        return _calculate_hash(hashlib.sha224(), file.absolute_path, CHUNK_SIZE)


class Crc32Tag(Tag):
    """CRC32 hash of the file"""

    require_context = False

    def process(self, file: File, context: Optional[str]) -> str:
        assert context is None
        hash_value = 0
        with open(file.absolute_path, "rb") as f:
            for chunk in iter(lambda: f.read(CHUNK_SIZE), b""):
                hash_value = zlib.crc32(chunk, hash_value)
        return f"{hash_value:08x}"



import fnmatch
import logging
import re
from abc import ABC, abstractmethod
from typing import Callable

from tempren.path_generator import (
    ExpressionEvaluationError,
    File,
    TemplateEvaluationError,
    evaluate_expression,
)
from tempren.template.tree_elements import Pattern


class FileFilter(ABC):
    @abstractmethod
    def __call__(self, file: File) -> bool:
        raise NotImplementedError()


class FileFilterInverter(FileFilter):
    def __init__(self, original_filter: Callable[[File], bool]):
        self.original_filter = original_filter

    def __call__(self, file: File) -> bool:
        return not self.original_filter(file)


class RegexFileFilter(FileFilter, ABC):
    pattern: re.Pattern
    ignore_case: bool

    def __init__(self, pattern: str, ignore_case: bool = False):
        flags = 0
        if ignore_case:
            flags |= re.IGNORECASE
        self.pattern = re.compile(pattern, flags)
        self.ignore_case = ignore_case


class RegexFilenameFileFilter(RegexFileFilter):
    def __call__(self, file: File) -> bool:
        match = self.pattern.match(file.relative_path.name)
        return match is not None


class RegexPathFileFilter(RegexFileFilter):
    def __call__(self, file: File) -> bool:
        match = self.pattern.match(str(file.relative_path))
        return match is not None


class GlobFileFilter(FileFilter, ABC):
    pattern: str
    ignore_case: bool

    def __init__(self, pattern: str, ignore_case: bool = False):
        self.pattern = pattern
        self.ignore_case = ignore_case
        if self.ignore_case:
            self.pattern = self.pattern.lower()


class GlobFilenameFileFilter(GlobFileFilter):
    def __call__(self, file: File) -> bool:
        if self.ignore_case:
            return fnmatch.fnmatch(file.relative_path.name.lower(), self.pattern)
        return fnmatch.fnmatchcase(file.relative_path.name, self.pattern)


class GlobPathFileFilter(GlobFileFilter):
    def __call__(self, file: File) -> bool:
        if self.ignore_case:
            return fnmatch.fnmatch(str(file.relative_path).lower(), self.pattern)
        return fnmatch.fnmatchcase(str(file.relative_path), self.pattern)


class TemplateFileFilter(FileFilter):
    log: logging.Logger
    pattern: Pattern

    def __init__(self, pattern: Pattern):
        self.log = logging.getLogger(__name__)
        self.pattern = pattern

    def __call__(self, file: File) -> bool:
        self.log.debug("Rendering filter template for '%s'", file)
        rendered_expression = self.pattern.process_as_expression(file)
        self.log.debug("Evaluating filter expression '%s'", rendered_expression)
        try:
            evaluation_result = evaluate_expression(rendered_expression)
            self.log.debug("Evaluation result '%s'", repr(evaluation_result))
            return bool(evaluation_result)
        except ExpressionEvaluationError as evaluation_error:
            assert self.pattern.source_representation is not None
            raise TemplateEvaluationError(
                file,
                self.pattern.source_representation,
                evaluation_error.expression,
                evaluation_error.message,
            )



import logging
from abc import ABC, abstractmethod
from typing import Iterable, Tuple

from tempren.path_generator import (
    ExpressionEvaluationError,
    File,
    TemplateEvaluationError,
    evaluate_expression,
)
from tempren.template.tree_elements import Pattern


class FileSorter(ABC):
    @abstractmethod
    def __call__(self, files: Iterable[File]) -> Iterable[File]:
        raise NotImplementedError()


class TemplateFileSorter(FileSorter):
    log: logging.Logger
    pattern: Pattern
    invert: bool = False

    def __init__(self, pattern: Pattern, invert: bool = False):
        self.log = logging.getLogger(__name__)

        self.pattern = pattern
        self.invert = invert

    def __call__(self, files: Iterable[File]) -> Iterable[File]:
        return sorted(files, key=self._generate_sort_key, reverse=self.invert)

    def _generate_sort_key(self, file: File) -> Tuple:
        self.log.debug("Rendering sorting value template for '%s'", file)
        rendered_expression = "(" + self.pattern.process_as_expression(file) + ", )"
        self.log.debug("Evaluating sorting value expression '%s'", rendered_expression)
        try:
            evaluation_result = evaluate_expression(rendered_expression)
            self.log.debug("Evaluation result '%s'", repr(evaluation_result))
            return evaluation_result
        except ExpressionEvaluationError as evaluation_error:
            assert self.pattern.source_representation is not None
            raise TemplateEvaluationError(
                file,
                self.pattern.source_representation,
                evaluation_error.expression,
                evaluation_error.message,
            )



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['tempren', 'tempren.tags', 'tempren.template', 'tempren.template.grammar']

package_data = \
{'': ['*']}

install_requires = \
['Pillow>=9.2.0,<10.0.0',
 'Unidecode>=1.2.0,<2.0.0',
 'antlr4-python3-runtime>=4.10,<5.0',
 'docstring-parser>=0.13,<0.14',
 'isodate>=0.6.1,<0.7.0',
 'mutagen>=1.45.1,<2.0.0',
 'pathvalidate>=2.4.1,<3.0.0',
 'piexif>=1.1.3,<2.0.0',
 'pymediainfo>=5.1.0,<6.0.0',
 'python-magic>=0.4.27,<0.5.0']

entry_points = \
{'console_scripts': ['tempren = tempren.cli:throwing_main']}

setup_kwargs = {
    'name': 'tempren',
    'version': '0.6.0',
    'description': 'Template-based renaming utility',
    'long_description': '# Tempren - template-based file renaming utility\n\n![run-tests](https://github.com/idle-code/tempren/actions/workflows/run-tests.yml/badge.svg)\n[![codecov](https://codecov.io/gh/idle-code/tempren/branch/develop/graph/badge.svg?token=1CR2PX6GYB)](https://codecov.io/gh/idle-code/tempren)\n[![PyPI version](https://badge.fury.io/py/tempren.svg)](https://badge.fury.io/py/tempren)\n\n`tempren` is a powerful batch file renamer that can generate filenames based on flexible template expressions.\nIt can create new, modify existing filenames or sort files into directories based on their attributes (metadata).\nRich library of build-in tags allow for easy metadata extraction from common file types.\n\n## Features\n- Template-based filename/path generation\n- Configurable file selection (filtering)\n- Metadata-based sorting\n- Multimedia metadata extraction\n\n## Installation\nCurrently only PyPI installation is supported, just run following command:\n```commandline\n$ pip install [--user] tempren\n```\n\n## [Quickstart](QUICKSTART.md)\nFor quick, five-minute introduction to the most of `tempren` features please refer to the [quickstart](QUICKSTART.md) page.\nYou can also take a look on the following examples.\n\n## Examples\n**Note: When experimenting on your own please use `--dry-run` flag!** \\\n**Tempren will not override your files by default but invalid template can mess them up a bit.**\n\nCleaning up names for sensitive (e.g. FAT32) filesystems:\n```commandline\n$ tempren --recursive --name "%Strip(){%Base()|%Unidecode()|%Sanitize()|%Collapse()}%Ext()" ./Some\\ OST/\nRenamed: Disk 1/14 - 接近.flac\n     to: Disk 1/14 - Jie Jin.flac\nRenamed: Disk 1/02 - なつのあお.flac\n     to: Disk 1/02 - natsunoao.flac\nRenamed: Disk 1/11 - 灯火-re.flac\n     to: Disk 1/11 - Deng Huo -re.flac\nRenamed: Disk 1/05 - 記録.flac\n     to: Disk 1/05 - Ji Lu.flac\nRenamed: Disk 1/10 - むかしむかし、あるところに.flac\n     to: Disk 1/10 - mukashimukashi, arutokoroni.flac\nRenamed: Disk 1/09 - 阿良句のテーマ(ハイ).flac\n     to: Disk 1/09 - A Liang Ju notema(hai).flac\n...\n```\n\nAdding resolution to the image files:\n```commandline\n$ tempren --name "%Base()_%Image.Width()x%Image.Height()%Ext()" ~/Pictures/Wallpapers\nRenamed: 0sa5yfiskqr21.jpg\n     to: 0sa5yfiskqr21_3728x4660.jpg\nRenamed: rkgjq6883fp81.jpg\n     to: rkgjq6883fp81_3024x4032.jpg\nRenamed: lcrkvphf28911.jpg\n     to: lcrkvphf28911_4016x4684.jpg\nRenamed: y6nzcv55k3851.jpg\n     to: y6nzcv55k3851_3784x5670.jpg\nRenamed: 1211740803547.jpg\n     to: 1211740803547_1200x1109.jpg\n...\n```\n\nSorting files into directories based on their MIME type:\n```commandline\n$ tempren -d --path "%Capitalize(){%Mime(subtype)}/%Name()" ~/Downloads\nRenamed: dotnet-install.sh\n     to: X-shellscript/dotnet-install.sh\nRenamed: openrgb_0.7_amd64_buster_6128731.deb\n     to: Vnd.debian.binary-package/openrgb_0.7_amd64_buster_6128731.deb\nRenamed: prometheus-2.26.0.linux-amd64.tar.gz\n     to: Gzip/prometheus-2.26.0.linux-amd64.tar.gz\nRenamed: nldb remote.zip\n     to: Zip/nldb remote.zip\nRenamed: artifacts.zip\n     to: Zip/artifacts.zip\nRenamed: 2021-06-11_12-09-34.webm\n     to: X-matroska/2021-06-11_12-09-34.webm\nRenamed: antlr-4.9.2-complete.jar\n     to: Java-archive/antlr-4.9.2-complete.jar\n...\n```\n\nAdding checksums to the names of the audio files:\n```commandline\n$ tempren --filter-template "%IsMime(\'audio\')" --name "%Base() [%Upper(){%Crc32()}]%Ext()" ./Roger\\ Subirana\\ Mata\\ -\\ Point\\ of\\ no\\ return\nRenamed: 10-169205-Roger Subirana Mata-Island of light.mp3\n     to: 10-169205-Roger Subirana Mata-Island of light [08E46C33].mp3\nRenamed: 12-169207-Roger Subirana Mata-Tales of trees.mp3\n     to: 12-169207-Roger Subirana Mata-Tales of trees [33EFEC5E].mp3\nRenamed: 11-169206-Roger Subirana Mata-Requiem.mp3\n     to: 11-169206-Roger Subirana Mata-Requiem [5E48759B].mp3\nRenamed: 05-168950-Roger Subirana Mata-The mask.mp3\n     to: 05-168950-Roger Subirana Mata-The mask [045DBC19].mp3\nRenamed: 03-168948-Roger Subirana Mata-Thryst.mp3\n     to: 03-168948-Roger Subirana Mata-Thryst [5D23E43B].mp3\n...\n```\n\n## Contributing\n\nIf you noticed a bug please or have and idea for new tag please open an issue with appropriate tags.\nIf you would like to contribute to the development please visit [contributing page](CONTRIBUTING.md).\n',
    'author': 'Paweł Żukowski',
    'author_email': 'p.z.idlecode@gmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/idle-code/tempren',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



#!/usr/bin/env python
import argparse
import logging
import sys
from argparse import ArgumentParser, Namespace
from enum import IntEnum
from logging import LogRecord
from pathlib import Path
from textwrap import indent
from typing import Any, List, NoReturn, Optional, Sequence, Text, Union

from tempren.filesystem import DestinationAlreadyExistsError
from tempren.path_generator import TemplateEvaluationError
from tempren.template.tree_elements import TagName

from .pipeline import (
    ConflictResolutionStrategy,
    FilterType,
    InvalidDestinationError,
    OperationMode,
    RuntimeConfiguration,
    build_pipeline,
    build_tag_registry,
)
from .template.tree_builder import (
    AmbiguousTagError,
    TagError,
    TemplateError,
    UnknownCategoryError,
    UnknownTagError,
)

log = logging.getLogger("CLI")


class LogFormatter(logging.Formatter):
    def __init__(self) -> None:
        super().__init__()

    def format(self, record: LogRecord) -> str:
        if logging.root.level <= logging.NOTSET:  # NOCOVER: developer option
            return f"{record.name}:{record.levelname}: {record.getMessage()}"
        if record.levelno == logging.INFO:
            return record.getMessage()
        else:
            return f"{record.levelname}: {record.getMessage()}"


class LogLevelFilter(logging.Filter):
    def __init__(self, min_level: int, max_level: int):
        super().__init__()
        assert min_level <= max_level
        self.min_level = min_level
        self.max_level = max_level

    def filter(self, record: LogRecord) -> bool:
        return self.min_level <= record.levelno <= self.max_level


# TODO: Use dictConfig for more readability
# CHECK: Move to config module?
def configure_logging():
    stdout_handler = logging.StreamHandler(sys.stdout)
    stdout_handler.addFilter(LogLevelFilter(0, logging.INFO))
    stdout_handler.setFormatter(LogFormatter())
    stderr_handler = logging.StreamHandler(sys.stderr)
    stderr_handler.addFilter(LogLevelFilter(logging.WARNING, 1000))
    stderr_handler.setFormatter(LogFormatter())
    logging.root.setLevel(logging.INFO)
    logging.root.addHandler(stdout_handler)
    logging.root.addHandler(stderr_handler)


def existing_directory(val: str) -> Path:
    directory_path = Path(val)
    if not directory_path.is_dir():
        raise argparse.ArgumentTypeError(f"Directory '{val}' doesn't exists")
    return directory_path


def nonempty_string(val: str) -> str:
    if not val:
        raise argparse.ArgumentTypeError(f"Non-empty argument required")
    return val


class SystemExitError(Exception):
    status: int

    def __init__(self, status: int, message: Optional[str]):
        super().__init__(message)
        self.status = status


class _ListAvailableTags(argparse.Action):
    def __call__(
        self,
        parser: ArgumentParser,
        namespace: Namespace,
        values: Union[Text, Sequence[Any], None],
        option_string: Optional[Text] = None,
    ):
        registry = build_tag_registry()
        log.info("Available tags:")
        for category_name in sorted(registry.category_map.keys()):
            log.info(f"{category_name.capitalize()}:")
            category = registry.category_map[category_name]

            all_category_tags = sorted(category.tag_map.items())
            if not all_category_tags:
                # There is no tags in the category
                # TODO: Empty categories should not be present in the registry
                continue
            max_name_length = max(
                [len(tag_name) for tag_name, factory in all_category_tags]
            )
            log.debug(
                "Longest tag name: %d in category %s", max_name_length, category_name
            )
            for tag_name, factory in all_category_tags:
                log.info(
                    f"  {tag_name.ljust(max_name_length)} - {factory.short_description}"
                )
        parser.exit()


class _ShowHelp(argparse.Action):
    def __call__(
        self,
        parser: ArgumentParser,
        namespace: Namespace,
        values: Union[str, Sequence[Any], None],
        option_string: Optional[Text] = None,
    ):
        if values is None:
            parser.print_help()
        else:
            raw_tag_name = str(values)
            registry = build_tag_registry()

            try:
                if "." in raw_tag_name:
                    tag_name = TagName(*reversed(raw_tag_name.split(".")))
                else:
                    tag_name = TagName(raw_tag_name)

                tag_factory = registry.get_tag_factory(tag_name)
            except TagError as tag_error:
                parser.exit(ErrorCode.USAGE_ERROR, str(tag_error))
                return
            log.info(tag_factory.configuration_signature)
            log.info("")
            log.info(tag_factory.short_description)
            if tag_factory.long_description:
                log.info("")
                log.info(tag_factory.long_description)

        parser.exit()


class _ShowVersion(argparse.Action):
    def __call__(
        self,
        parser: ArgumentParser,
        namespace: Namespace,
        values: Union[Text, Sequence[Any], None],
        option_string: Optional[Text] = None,
    ):
        log.info(self.find_package_version())
        parser.exit()

    @staticmethod
    def find_package_version() -> str:  # NOCOVER: hard to test - not worth it
        package_name = "tempren"
        try:
            import importlib.metadata

            return importlib.metadata.version(package_name)
        except ModuleNotFoundError:
            try:
                import importlib_metadata

                return importlib_metadata.version(package_name)
            except ModuleNotFoundError:
                return "0.0.0"


class _IncreaseLogVerbosity(argparse.Action):
    def __call__(
        self,
        parser: ArgumentParser,
        namespace: Namespace,
        values: Union[Text, Sequence[Any], None],
        option_string: Optional[Text] = None,
    ):
        root_logger = logging.getLogger()
        root_logger.setLevel(root_logger.level - 10)
        log.debug("Verbosity level set to %d", root_logger.level)


class _DecreaseLogVerbosity(argparse.Action):
    def __call__(
        self,
        parser: ArgumentParser,
        namespace: Namespace,
        values: Union[Text, Sequence[Any], None],
        option_string: Optional[Text] = None,
    ):
        root_logger = logging.getLogger()
        root_logger.setLevel(root_logger.level + 10)
        log.debug("Verbosity level set to %d", root_logger.level)


class DefaultOptionsHelpFormatter(argparse.HelpFormatter):
    """Indicates default options in help page"""

    def _get_help_string(self, action):
        if action.default:
            return "(default) " + action.help
        return action.help


# CHECK: use pydantic-cli for argument parsing
def process_cli_configuration(argv: List[str]) -> RuntimeConfiguration:
    log.debug("Parsing command line arguments")
    parser = argparse.ArgumentParser(
        prog="tempren",
        description="Template-based renaming utility",
        fromfile_prefix_chars="@",
        add_help=False,
        formatter_class=DefaultOptionsHelpFormatter,
    )

    parser.add_argument(
        "-d",
        "--dry-run",
        action="store_true",
        help="Do not perform any renaming - just show expected operation results",
    )
    parser.add_argument(
        "-r",
        "--recursive",
        action="store_true",
        help="Look for files in input directory recursively",
    )
    parser.add_argument(
        "-ih",
        "--include-hidden",
        action="store_true",
        help="Consider hidden files and directories when scanning for files in input directory",
    )
    parser.add_argument(
        "-v",
        "--verbose",
        action=_IncreaseLogVerbosity,
        nargs=0,
        help="Increase logging verbosity",
    )
    parser.add_argument(
        "-q",
        "--quiet",
        action=_DecreaseLogVerbosity,
        nargs=0,
        help="Decrease logging verbosity",
    )

    operation_modes_group = parser.add_argument_group("operation modes")
    operation_mode = operation_modes_group.add_mutually_exclusive_group()
    operation_mode.add_argument(
        "-n",
        "--name",
        action="store_const",
        dest="mode",
        const=OperationMode.name,
        default=OperationMode.name,
        help="Use template to generate file name",
    )
    operation_mode.add_argument(
        "-p",
        "--path",
        action="store_const",
        dest="mode",
        const=OperationMode.path,
        help="Use template to generate relative file path",
    )

    conflict_resolution_group = parser.add_argument_group("conflict resolution")
    conflict_resolution = conflict_resolution_group.add_mutually_exclusive_group()
    conflict_resolution.add_argument(
        "-cs",
        "--conflict-stop",
        action="store_true",
        default=True,
        help="Keep source file name intact and stop",
    )
    conflict_resolution.add_argument(
        "-ci",
        "--conflict-ignore",
        action="store_true",
        help="Keep source file name intact and continue",
    )
    conflict_resolution.add_argument(
        "-co",
        "--conflict-override",
        action="store_true",
        help="Override destination file",
    )
    conflict_resolution.add_argument(
        "-cm",
        "--conflict-manual",
        action="store_true",
        help="Prompt user to resolve conflict manually (choose an option or provide new filename)",
    )

    filtering_group = parser.add_argument_group("filtering")
    filter_mode = filtering_group.add_mutually_exclusive_group()
    filter_mode.add_argument(
        "-fg",
        "--filter-glob",
        type=str,
        metavar="filter_expression",
        help="Globbing filter expression to include individual files",
    )
    filter_mode.add_argument(
        "-fr",
        "--filter-regex",
        type=str,
        metavar="filter_expression",
        help="Regex filter expression to include individual files",
    )
    filter_mode.add_argument(
        "-ft",
        "--filter-template",
        type=nonempty_string,
        metavar="filter_expression",
        help="Tag template filter expression to include individual files",
    )
    filtering_group.add_argument(
        "-fi",
        "--filter-invert",
        action="store_true",
        help="Invert (negate) filter expression result",
    )

    sorting_group = parser.add_argument_group("sorting")
    sorting_group.add_argument(
        "-s",
        "--sort",
        type=nonempty_string,
        metavar="sort_expression",
        help="Sorting tag template used to order file list before processing",
    )
    sorting_group.add_argument(
        "-si",
        "--sort-invert",
        action="store_true",
        help="Reverse sorting order",
    )

    parser.add_argument(
        "template",
        type=nonempty_string,
        help="Template used to generate new filename/path",
    )
    parser.add_argument(
        "input_directory",
        type=existing_directory,
        help="Input directory where files to rename are stored",
    )
    parser.add_argument(
        "-l",
        "--list-tags",
        action=_ListAvailableTags,
        nargs=0,
        help="List available tags and exit",
    )
    parser.add_argument(
        "-h",
        "--help",
        action=_ShowHelp,
        type=str,
        nargs="?",
        metavar="tag_name",
        help="Show help message or tag documentation and exit",
    )
    parser.add_argument(
        "--version",
        action=_ShowVersion,
        nargs=0,
        help="Print version and exit",
    )

    # Upon parsing error, ArgumentParser tries to exit via calling `sys.exit()`.
    # We could catch resulting `SystemExit` exception but related error message still would be missing.
    # This behaviour is not comfortable for testing so here we monkeypatch exit method to
    # throw more practical exception instead.
    def throwing_exit(status: int = 0, message: Optional[str] = "") -> NoReturn:
        raise SystemExitError(status, message)

    parser.exit = throwing_exit  # type: ignore

    args = parser.parse_args(argv)

    if args.filter_glob:
        filter_type = FilterType.glob
        filter_expression = args.filter_glob
    elif args.filter_regex:
        filter_type = FilterType.regex
        filter_expression = args.filter_regex
    elif args.filter_template:
        filter_type = FilterType.template
        filter_expression = args.filter_template
    else:
        filter_type = FilterType.glob
        filter_expression = None

    # TODO: User action='store_const' (as with mode) in parser.add_argument to simplify this
    if args.conflict_ignore:
        conflict_strategy = ConflictResolutionStrategy.ignore
    elif args.conflict_override:
        conflict_strategy = ConflictResolutionStrategy.override
    elif args.conflict_manual:
        conflict_strategy = ConflictResolutionStrategy.manual
    elif args.conflict_stop:
        conflict_strategy = ConflictResolutionStrategy.stop
    else:
        raise NotImplementedError()

    configuration = RuntimeConfiguration(
        template=args.template,
        input_directory=args.input_directory,
        recursive=args.recursive,
        include_hidden=args.include_hidden,
        dry_run=args.dry_run,
        filter_type=filter_type,
        filter_invert=args.filter_invert,
        filter=filter_expression,
        conflict_strategy=conflict_strategy,
        sort_invert=args.sort_invert,
        sort=args.sort,
        mode=args.mode,
    )

    return configuration


def render_template_error(template_error: TemplateError, indent_size: int = 4):
    assert template_error.location.line == 1, "No support for multiline templates yet"
    if template_error.location.length == 1:
        underline = " " * template_error.location.column + "^"
    else:
        underline = (
            " " * template_error.location.column + "~" * template_error.location.length
        )
    log.error("")
    log.error(indent(template_error.template, " " * indent_size))
    log.error(indent(underline, " " * indent_size))
    log.error(f"Template error at {template_error.location}: {template_error.message}")


def render_template_evaluation_error(
    template_error: TemplateEvaluationError, indent_size: int = 4
):
    assert template_error.template
    assert template_error.expression

    log.error("While processing file %r with template:", template_error.file)
    log.error(indent(template_error.template, " " * indent_size))
    log.error("which has been rendered to:")
    log.error(indent(repr(template_error.expression), " " * indent_size))
    log.error("an evaluation error occurred:")
    log.error(indent(template_error.message, " " * indent_size))


def cli_prompt_conflict_resolver(
    source_path: Path, destination_path: Path
) -> Union[ConflictResolutionStrategy, Path]:
    log.warning("White processing:")
    log.warning(f"  {source_path}")
    log.warning("following path was generated:")
    log.warning(f"  {destination_path}")
    log.warning("but it cannot be used as it targets already existing file")
    while True:
        selected_option_name = input(
            "[s]top, [o]verride, [c]ustom path, [I]gnore: "
        ).lower()
        if not selected_option_name or "ignore".startswith(selected_option_name):
            return ConflictResolutionStrategy.ignore
        if "stop".startswith(selected_option_name):
            return ConflictResolutionStrategy.stop
        if "override".startswith(selected_option_name):
            return ConflictResolutionStrategy.override
        if "custom path".startswith(selected_option_name):
            custom_path = Path(input("Custom path: "))
            return custom_path

        log.error(f"Invalid choice '{selected_option_name}")


class ErrorCode(IntEnum):
    SUCCESS = 0
    """Everything went as it should have"""

    INVALID_DESTINATION_ERROR = 1
    """Something is wrong with generated destination path"""

    USAGE_ERROR = 2  # argparse uses 2 for its errors too
    """User tried to use program not as it should be used"""

    TEMPLATE_SYNTAX_ERROR = 3
    """Syntax error detected in one of provided templates"""

    TEMPLATE_EVALUATION_ERROR = 4
    """Python evaluation of filtering/sorting template failed"""

    UNKNOWN_ERROR = 126
    """This should not happen"""


def main() -> int:
    configure_logging()
    argv = sys.argv[1:]
    try:
        config = process_cli_configuration(argv)
        registry = build_tag_registry()
        pipeline = build_pipeline(
            config, registry, manual_conflict_resolver=cli_prompt_conflict_resolver
        )
        pipeline.execute()
        log.info("Done")
        return ErrorCode.SUCCESS
    # TODO: Clean up exception hierarchy
    except SystemExitError as exc:
        if exc.status != 0:
            log.error(exc)
        return exc.status
    except TemplateEvaluationError as template_evaluation_error:
        render_template_evaluation_error(template_evaluation_error)
        return ErrorCode.TEMPLATE_EVALUATION_ERROR
    except TemplateError as template_error:
        render_template_error(template_error)
        return ErrorCode.TEMPLATE_SYNTAX_ERROR
    except DestinationAlreadyExistsError as exc:
        log.error(f"Error: {exc}")
        return ErrorCode.INVALID_DESTINATION_ERROR
    except InvalidDestinationError as exc:
        log.error(f"Error: {exc}")
        return ErrorCode.INVALID_DESTINATION_ERROR
    except Exception as exc:
        log.error(f"Unknown error: {exc}")
        return ErrorCode.UNKNOWN_ERROR


def throwing_main():
    sys.exit(main())


if __name__ == "__main__":
    throwing_main()
